"""
Base classes for visualizing Helipad models with Matplotlib, to be passed to `model.useVisual()`. Main visualizers are `TimeSeries` to display diachronic data, and `Charts` to display synchronic data.
"""

import sys
from abc import ABC, abstractmethod
from math import sqrt, ceil, pi, isnan
from numpy import ndarray, array, asanyarray, log10, linspace, newaxis, arange, full_like, linalg, ones, vstack
import matplotlib, matplotlib.pyplot as plt, matplotlib.style as mlpstyle
from matplotlib.lines import Line2D
from matplotlib.patches import Polygon
from helipad.helpers import *
mlpstyle.use('fast')

#======================
# TOP-LEVEL VISUALIZERS
# These two use Matplotlib, but subclasses of BaseVisualization don't necessarily have to
#======================

class BaseVisualization(ABC):
	"""Base class defining the methods that must be implemented by any visualization. https://helipad.dev/functions/basevisualization/"""
	isNull: bool = True

	@abstractmethod
	def launch(self, title: str):
		"""Launch the visualization window (if in Tkinter) or cell (if in Jupyter). https://helipad.dev/functions/basevisualization/launch/"""

	@abstractmethod
	def refresh(self, data: dict):
		"""Update the visualization with new data and refresh the display. `data` is only data since the last visualization refresh. https://helipad.dev/functions/basevisualization/refresh/"""

	@abstractmethod
	def event(self, t: int, color, **kwargs):
		"""Called when an event is triggered, in order to be reflected in the visualization. https://helipad.dev/functions/basevisualization/event/"""

	def terminate(self, model):
		"""Cleanup on model termination. Called automatically from `model.terminate()`. https://helipad.dev/functions/basevisualization/terminate/"""

class MPLVisualization(BaseVisualization, dict):
	"""Base class for visualizations using Matplotlib. https://helipad.dev/functions/mplvisualization/"""

	def __init__(self, model):
		self.model = model #Unhappy with this
		self.selector = model.params.add('plots', ï('Plots'), 'checkgrid', [], opts={}, runtime=False, config=True)
		self.dim = None
		self.pos = (400, 0)
		self.fig = None
		self.lastUpdate = None
		self.keyListeners = {}

		def pause(model, event):
			if model.hasModel and event.canvas is self.fig.canvas:
				if model.running: model.stop()
				else: model.start()
		self.addKeypress(' ', pause)

		if isNotebook():
			from IPython import get_ipython
			get_ipython().magic('matplotlib widget')
		else: matplotlib.use('TkAgg') #macosx would be preferable (Retina support), but it blocks the cpanel while running

	def __repr__(self): return f'<{self.__class__.__name__} with {len(self)} plots>'

	# Subclasses should call super().launch **after** the figure is created.
	@abstractmethod
	def launch(self, title: str):
		if not isNotebook():
			self.fig.canvas.manager.set_window_title(title)
			if self.model.cpanel: self.model.cpanel.setAppIcon()

		#MPL can't take a method bound to an unhashable class (i.e. `dict`)
		def sendEvent(event):
			axes = event.artist.axes if hasattr(event, 'artist') else event.inaxes
			if axes is not None:
				for p in self.activePlots.values():
					if axes is p.axes:
						p.MPLEvent(event)
						break

			if event.name=='key_press_event' and event.key in self.keyListeners:
				for f in self.keyListeners[event.key]: f(self.model, event)

		self.fig.tight_layout()
		self.fig.canvas.mpl_connect('close_event', self.model.terminate)
		self.fig.canvas.mpl_connect('key_press_event', sendEvent)
		self.fig.canvas.mpl_connect('pick_event', sendEvent)
		self.fig.canvas.mpl_connect('button_press_event', sendEvent)
		self.lastUpdate = 0

		#Resize and position graph window if applicable
		fm = self.fig.canvas.manager
		if hasattr(fm, 'window'):
			if not self.dim:
				#This should be the window height, but MPL only allows us to set the figure height.
				#MacOS doesn't let us create a window taller than the screen, but we have to account for
				#the height of the window frame crudely in Windows.
				height = fm.window.winfo_screenheight()
				if sys.platform=='win32': height -= 150

				width = fm.window.winfo_screenwidth()*2/3
				width = min(width, fm.window.winfo_screenwidth()-400)
			else: width, height = self.dim

			self.fig.set_size_inches(width/self.fig.dpi, height/self.fig.dpi)
			fm.window.wm_geometry(f'+{self.pos[0]}+{self.pos[1]}')

	#Save window dimensions and position
	def terminate(self, model):
		if self.isNull: return
		fm = self.fig.canvas.manager
		if hasattr(fm, 'window'):
			self.dim = list(self.fig.get_size_inches()*self.fig.dpi)
			pos = fm.window.wm_geometry().split('+')
			self.pos = (pos[1], pos[2])

	def addKeypress(self, key: str, fn):
		"""Register a function to be run when `key` is pressed in a Matplotlib visualizer. `fn` will run if `key` is pressed at any time when the plot window is in focus. To narrow the focus to a particular plot, define `catchKeypress()` in a subclass of `ChartPlot`. https://helipad.dev/functions/mplvisualization/addkeypress/"""
		if key not in self.keyListeners: self.keyListeners[key] = []
		self.keyListeners[key].append(fn)

	@property
	def activePlots(self) -> dict:
		"""The subset of the plots containing the `Plot`s that are currently active. https://helipad.dev/functions/mplvisualization/#activePlots"""
		return {k:plot for k,plot in self.items() if plot.selected}

	@property
	def isNull(self) -> bool:
		"""`True` when the model should run as-if with no visualization, for example if all plots are unselected. `False` indicates the window can be launched. https://helipad.dev/functions/mplvisualization/#isNull"""
		return not [plot for plot in self.values() if plot.selected]

	#Deprecated in Helipad 1.6, remove in 1.8
	@property
	def plots(self):
		"""A `dict` of plots. This property is deprecated; the visualization object can be indexed directly."""
		warnings.warn(ï('model.visual.plots is deprecated. Plots can be accessed by indexing the visualization object directly.'), FutureWarning, 2)
		return self

class TimeSeries(MPLVisualization):
	"""A Matplotlib-based visualizer for displaying time series data on plots so the whole history of any given variable over the model's runtime can be seen at once. https://helipad.dev/functions/timeseries/"""
	def __init__(self, model):
		super().__init__(model)
		self.verticals = []

		#Plot categories
		self.addPlot('utility', 'Utility', selected=False)

		if len(model.goods) >= 2: self.addPlot('demand', ï('Demand'), selected=False)
		if model.goods.money is not None: self.addPlot('money', ï('Money'), selected=False)

		#Toggle legend boxes all at once
		#Use event.canvas.figure and not self.fig so it pertains to the current window
		def toggle(model, event):
			for axes in event.canvas.figure.axes:
				leg = axes.get_legend()
				leg.set_visible(not leg.get_visible())
			event.canvas.draw_idle()
		self.addKeypress('t', toggle)

		#Delete the corresponding series when a reporter is removed
		@model.hook('removeReporter')
		def deleteSeries(data, key):
			for p in self.values():
				for s in p.series:
					if s.reporter==key:
						# Remove subseries
						for ss in s.subseries:
							for sss in self[s.plot].series:
								if sss.reporter == ss:
									self[s.plot].series.remove(sss)
									continue
						self[s.plot].series.remove(s)

		#Move the plots parameter to the end when the cpanel launches
		@model.hook('CpanelPreLaunch')
		def movePlotParam(model):
			model.params['plots'] = model.params.pop('plots')

	def launch(self, title: str):
		if not self.activePlots: return #Windowless mode

		self.resolution = 1
		if isNotebook():
			plt.close() #Clean up after any previous runs
			matplotlib.rcParams['figure.figsize'] = [9, 7]

		#fig is the figure, plots is a list of AxesSubplot objects
		#The Tkinter way of setting the title doesn't work in Jupyter
		#The Jupyter way works in Tkinter, but also serves as the figure id, so new graphs draw on top of old ones
		self.fig, plots = plt.subplots(len(self.activePlots), sharex=True, num=title if isNotebook() else None)
		super().launch(title)

		if not isinstance(plots, ndarray): plots = asanyarray([plots]) #.subplots() tries to be clever & returns a different data type if len(plots)==1
		for plot, axes in zip(self.activePlots.values(), plots): plot.launch(axes)

		#Style plots
		self.fig.subplots_adjust(hspace=0, bottom=0.05, right=1, top=0.97, left=0.1)
		# plt.setp([a.get_xticklabels() for a in self.fig.axes[:-1]], visible=False)	#What was this for again…?

		self.fig.canvas.draw_idle()
		plt.show(block=False)

	def terminate(self, model):
		super().terminate(model)
		if self.isNull:
			model.params['csv'].enable()
			if not isinstance(model.param('stopafter'), str): model.params['stopafter'].enable()

	def refresh(self, data: dict):
		newlen = len(next(data[x] for x in data))
		if self.resolution > 1: data = {k: keepEvery(v, self.resolution) for k,v in data.items()}
		time = newlen + len(next(iter(self.activePlots.values())).series[0].fdata)*self.resolution
		for plot in self.activePlots.values(): plot.update(data, time) #Append new data to series

		#Redo resolution at 2500, 25000, etc
		if 10**(log10(time/2.5)-3) >= self.resolution:
			self.resolution *= 10
			for plot in self.activePlots.values():
				for serie in plot.series:
					serie.fdata = keepEvery(serie.fdata, 10)
			if self.resolution > self.model.param('refresh'): self.model.param('refresh', self.resolution)

		for plot in self.activePlots.values(): plot.draw(time) #Update the actual plots. Has to be after the new resolution is set
		if self.fig.stale: self.fig.canvas.draw_idle()
		self.fig.canvas.flush_events() #Listen for user input

	def addPlot(self, name: str, label: str, position=None, selected: bool=True, logscale: bool=False, stack: bool=False):
		"""Register a plot area in the `TimeSeries` plot area to which data series can be added. `position` Position is the number you want it to be, *not* the array position. https://helipad.dev/functions/timeseries/addplot/"""
		plot = TimeSeriesPlot(viz=self, name=name, label=label, logscale=logscale, stack=stack)
		if name in self: warnings.warn(ï("Replacing plot '{}'").format(name), None, 2)

		self.selector.addItem(name, label, position, selected)
		if position is None or position > len(self): self[name] = plot
		else:		#Reconstruct the dicts because there's no insert method…
			newplots, i = ({}, 1)
			for k,v in self.items():
				if position==i:
					newplots[name] = plot
				newplots[k] = v
				i+=1
			self.clear()
			self.update(newplots)

		plot.selected = selected #Do this after CheckgridParam.addItem
		return plot

	def removePlot(self, name, reassign=None):
		"""Remove a plot or plots and optionally reassign their series to a different plot. https://helipad.dev/functions/timeseries/removeplot/"""
		if self.model.cpanel: raise RuntimeError(ï('Cannot remove plots after control panel is drawn.'))
		if isinstance(name, list):
			for p in name: self.removePlot(p, reassign)
			return

		if name not in self:
			warnings.warn(ï('No plot \'{}\' to remove.').format(name), None, 2)
			return False

		if reassign is not None: self[reassign].series += self[name].series
		del self[name]
		del self.selector.opts[name]
		del self.selector.vars[name]
		if name in self.selector.default: self.selector.default.remove(name)
		return True

	def event(self, t: int, color='#CC0000', linestyle: str='--', linewidth=1, **kwargs):
		if self.fig is None: return
		self.verticals.append([p.axes.axvline(x=t, color=color, linestyle=linestyle, linewidth=linewidth) for p in self.activePlots.values()])

		# Problem: Need x to be in plot coordinates but y to be absolute w.r.t the figure
		# next(iter(self.values())).axes.text(t, 0, label, horizontalalignment='center')

class Charts(MPLVisualization):
	"""A Matplotlib-based visualizer for a variety of visualizations that display data that reflects a single point in time. https://helipad.dev/functions/charts/"""
	def __init__(self, model):
		super().__init__(model)
		self.events = {}
		self.plotTypes = {}
		self.timeslider = None

		for p in [BarChart, AgentsPlot, TimeSeriesPlot]: self.addPlotType(p)
		model.params['refresh'].runtime=False
		self.model = model # :(

	def launch(self, title: str):
		if not self.activePlots: return #Windowless mode
		from matplotlib.widgets import Slider

		n = len(self.activePlots)
		x = ceil(sqrt(n))
		y = ceil(n/x)
		if isNotebook():
			plt.close() #Clean up after any previous runs
			matplotlib.rcParams['figure.figsize'] = [9, 7]

		#Add the subplots individually rather than using plt.subplots() so we can mix and match projections
		self.fig = plt.figure()
		plots = list(self.activePlots.values())
		for i in range(n):
			plots[i].launch(self.fig.add_subplot(y,x,i+1, projection=plots[i].projection))
			plots[i].subplot_position = (y,x,i+1)
		super().launch(title)

		#Time slider
		ref = self.model.params['refresh'].get()
		self.fig.subplots_adjust(bottom=0.12) #Make room for the slider
		sax = self.fig.add_axes([0.1,0.01,.75,0.03], facecolor='#EEF')
		self.timeslider = Slider(sax, 't=', 0, ref, valinit=ref, valstep=ref, closedmin=False)
		self.timeslider.on_changed(self.scrub)

		self.fig.canvas.draw_idle()
		plt.show(block=False)

	def refresh(self, data: dict):
		data = {k:v[-1] for k,v in data.items()}
		t = self.model.t #cheating?
		for c in self.activePlots.values(): c.update(data, t)

		#Update slider. This calls self.scrub(), which in turn calls chart.draw()
		self.timeslider.valmax = t
		self.timeslider.set_val(t)
		self.timeslider.ax.set_xlim(0,t) #Refresh

		self.fig.patch.set_facecolor(self.events[t] if t in self.events else 'white')
		if self.fig.stale: self.fig.canvas.draw_idle()
		self.fig.canvas.flush_events() #Listen for user input

	def scrub(self, t: int):
		"""Update the visualizer with the values from model time `t`."""
		self.scrubval = t
		for c in self.activePlots.values(): c.draw(t)
		self.fig.patch.set_facecolor(self.events[t] if t in self.events else 'white')

	def addPlot(self, name: str, label: str, type=None, position=None, selected=True, **kwargs):
		"""Adds a plot area to the Chart visualizer. Defaults to a bar chart, but can be set to any subclass of ChartPlot. Kwargs are passed to the `ChartPlot` object. https://helipad.dev/functions/charts/addplot/"""
		if name in self: warnings.warn(ï("Replacing plot '{}'").format(name), None, 2)
		self.selector.addItem(name, label, position, selected)
		if type == 'network': #Deprecated in Helipad 1.6; remove in 1.8
			warnings.warn(ï('The `network` plot type is deprecated. Use `agents` instead.'), FutureWarning, 2)
			type = 'agents'
		self.type = type if type is not None else 'bar'
		if self.type not in self.plotTypes: raise KeyError(ï('\'{}\' is not a registered plot visualizer.').format(self.type))
		self[name] = self.plotTypes[self.type](name=name, label=label, viz=self, selected=True, **kwargs)

		self[name].selected = selected #Do this after CheckgridParam.addItem
		return self[name]

	def addPlotType(self, clss):
		"""Registers a new plot type for the Charts visualizer. Registered plot types can then be added to the visualization area with `Charts.addPlot()`. https://helipad.dev/functions/charts/addplottype/"""
		if not issubclass(clss, ChartPlot): raise TypeError(ï('New plot types must subclass ChartPlot.'))
		self.plotTypes[clss.type] = clss

	def removePlot(self, name):
		"""Remove a plot or plots and optionally reassign their series to a different plot. https://helipad.dev/functions/timeseries/removeplot/"""
		if self.model.cpanel: raise RuntimeError(ï('Cannot remove plots after control panel is drawn.'))
		if isinstance(name, list):
			for p in name: self.removePlot(p)
			return

		if name not in self:
			warnings.warn(ï('No plot \'{}\' to remove.').format(name), None, 2)
			return False

		del self[name]
		return True

	def event(self, t: int, color='#FDC', **kwargs):
		ref = self.model.params['refresh'].get()
		self.events[ceil(t/ref)*ref] = color

#======================
# PLOT-LEVEL VISUALIZERS
# Plug into either TimeSeries or Charts as one of the subplots
#======================

class ChartPlot(Item):
	"""Base class for creating a synchronic plot area in the Charts visualizer. Must interface with Matplotlib and specify `class.type`. Extra kwargs in `Charts.addPlot()` are passed to `ChartPlot.__init__()`. https://helipad.dev/functions/chartplot/"""
	def __init__(self, **kwargs):
		if 'projection' not in kwargs and not hasattr(self, 'projection'): self.projection = None
		self.axes = None
		super().__init__(**kwargs)

	def __repr__(self): return f'<{self.__class__.__name__}: {self.name}>'

	@property
	def selected(self):
		"""Whether the plot is currently queued for display. Do not modify this property directly; use ChartPlot.active() to ensure consistency with the GUI state. https://helipad.dev/functions/chartplot/#selected"""
		return self.viz.model.params['plots'].get(self.name)

	@selected.setter
	def selected(self, val: bool): self.active(val)

	def active(self, val: bool, updateGUI: bool=True):
		"""Turn on or off the selected state of a plot in the control panel. https://helipad.dev/functions/chartplot/active/"""
		self.viz.model.params['plots'].set(self.name, bool(val))
		if updateGUI and not isNotebook() and hasattr(self, 'check'):
			self.check.set(val)

	@abstractmethod
	def launch(self, axes):
		"""Set up the plot's `axes` object when `Charts` launches the visualization window. Subclasses should call `super().launch(axes)` at the beginning of the method. https://helipad.dev/functions/chartplot/launch/"""
		self.axes = axes
		axes.set_title(self.label, fontdict={'fontsize':10})

	@abstractmethod
	def update(self, data: dict, t: int):
		"""Receive current model data and store it for future scrubbing. Values of `data` should only contain the most recent value of each column. This method is only for updating plot data; drawing code happens in `draw()`. https://helipad.dev/functions/chartplot/update/"""

	@abstractmethod
	def draw(self, t: int, forceUpdate: bool=False):
		"""Refresh the `axes` object to display new or historical data. This method is called after `ChartPlot.update()` when the visualizer receives new data, and when scrubbing the time bar. https://helipad.dev/functions/chartplot/draw/"""
		if forceUpdate: self.viz.fig.canvas.draw_idle()

	def remove(self):
		"""Removes the plot from the visualizer."""
		self.viz.removePlot(self.name)

	def MPLEvent(self, event):
		"""Catch `pick_event`, `key_press_event`, and `button_press_event` events that occur inside a particular plot. https://helipad.dev/functions/chartplot/mplevent/"""

class TimeSeriesPlot(ChartPlot):
	"""Visualize time series data on one or more variables. Can be used in either the `TimeSeries` or `Charts` visualizer. https://helipad.dev/functions/timeseriesplot/"""
	type = 'timeseries'
	def __init__(self, **kwargs):
		self.series = []
		self.scrubline = None
		for p in ['logscale', 'stack']:
			if p not in kwargs: kwargs[p] = False
		super().__init__(**kwargs)

	def addSeries(self, reporter, label: str, color, style: str='-', visible: bool=True):
		"""Register a reporter to be plotted during the model's runtime. https://helipad.dev/functions/timeseriesplot/addseries/"""
		if not isinstance(color, Color): color = Color(color)

		#Check against columns and not reporters so subseries work
		if not callable(reporter) and reporter not in self.viz.model.data.columns:
			raise KeyError(ï('Reporter \'{}\' does not exist. Be sure to register reporters before adding series.').format(reporter))

		#Add subsidiary series (e.g. percentile bars)
		subseries = []
		if reporter in self.viz.model.data.reporters and self.viz.model.data.reporters[reporter].children:
			for p in self.viz.model.data.reporters[reporter].children:
				if '-unsmooth' in p: continue #Don't plot the unsmoothed series
				subseries.append(self.addSeries(p, '', color.lighten(), style='--'))

		#Since many series are added at setup time, we have to de-dupe
		for s in self.series:
			if s.reporter == reporter:
				self.series.remove(s)

		series = Series(reporter=reporter, label=label, color=color, style=style, subseries=subseries, plot=self.name, visible=visible)
		self.series.append(series)
		return series

	def launch(self, axes):
		self.axes = axes
		axes.margins(x=0)
		if self.stack:
			lines = axes.stackplot([], *[[] for s in self.series], color=[s.color.hex for s in self.series])
			for series, poly in zip(self.series, lines): series.poly = poly
			axes.margins(y=0)

		if self.logscale:
			axes.set_yscale('log')
			axes.set_ylim(1/2, 2, auto=True)

		#Create a line for each series
		#Do this even for stackplots because the line object is necessary to create the legend
		for series in self.series:
			series.line, = axes.plot([], label=series.label, color=series.color.hex, linestyle=series.style)
			series.fdata = []

		#Set up the legend for click events on both the line and the legend
		leg = axes.legend(loc='upper right')
		for legline, label in zip(leg.get_lines(), leg.get_texts()):
			legline.set_picker(True)	#Listen for mouse events on the legend line
			legline.set_pickradius(5)	#Set margin of valid events in pixels
			label.set_picker(5)			#Do both on the legend text
			for s in self.series:
				if s.label==label.get_text():
					s.legline = legline
					s.legtext = label
					label.axes = axes #Necessary because an MPL bug fails to set the axes property of the text labels. Fixed in 3.5.0 (#20458)
					break

		for series in self.series: #Make sure we start out with the user-set visibility
			if series.label and not series._visible: series.visible = False

	def update(self, data: dict, t: str):
		firstdata = next(iter(data.values()))
		if isinstance(firstdata, list): newlen, res = len(firstdata), self.resolution
		else:
			newlen = 1
			res = self.viz.model.param('refresh')
			data = {k: [v] for k,v in data.items()}
		for serie in self.series:
			if callable(serie.reporter):						#Lambda functions
				for i in range(newlen):
					serie.fdata.append(serie.reporter(t-(newlen-1-i)*res))
			elif serie.reporter in data: serie.fdata += data[serie.reporter] #Actual data
			else: continue									#No data

	def draw(self, t: int, forceUpdate: bool=False):
		if self.scrubline is not None:
			self.scrubline.remove()
			self.scrubline = None

		#Draw new data if this is a new refresh
		if t==len(self.series[0].fdata)*self.resolution:
			tseries = range(0, t, self.resolution)

			#No way to update the stack (?) so redraw it from scratch
			if self.stack:
				lines = self.axes.stackplot(tseries, *[s.fdata for s in self.series], colors=[s.color.hex for s in self.series])
				for series, poly in zip(self.series, lines): series.poly = poly
			else:
				for serie in self.series:
					serie.line.set_ydata(serie.fdata)
					serie.line.set_xdata(tseries)
				self.axes.relim()
				self.axes.autoscale_view(tight=False)

			#Prevent decaying averages on logscale graphs from compressing the entire view
			ylim = self.axes.get_ylim()
			if self.axes.get_yscale() == 'log' and ylim[0] < 10**-6: self.axes.set_ylim(bottom=10**-6)
			super().draw(t, forceUpdate)

		#Note the position in the old data if we're scrubbing
		else:
			self.scrubline = self.axes.axvline(x=t, color='#330000')

	def MPLEvent(self, event):
		#Toggle legend boxes one at a time if we're in the Charts visualizer
		if hasattr(self.viz, 'scrub') and event.name=='key_press_event' and event.key=='t':
			leg = self.axes.get_legend()
			leg.set_visible(not leg.get_visible())
			event.canvas.draw_idle()

		#Toggle lines on and off when clicking the legend
		if event.name=='pick_event':
			c1 = event.artist					#The label or line that was clicked
			for s in self.series:
				if s.label and (c1 is s.legline or c1 is s.legtext):
					s.toggle()
					break

	@property
	def resolution(self) -> int:
		"""The time elapsing between any two data points drawn to screen (not the time elapsing between two updates, which is a user-controlled parameter)."""
		return self.viz.resolution if hasattr(self.viz, 'resolution') else int(self.viz.model.param('refresh'))

class BarChart(ChartPlot):
	"""A plot type that displays live-updating bar charts. https://helipad.dev/functions/barchart/"""
	type = 'bar'
	def __init__(self, **kwargs):
		for arg in ['horizontal', 'logscale']:
			if arg not in kwargs: kwargs[arg] = False
		super().__init__(**kwargs)
		self.bars = []

	def addBar(self, reporter, label: str, color='blue', position=None):
		"""Register a data reporter as a bar on the current plot. To display error bars, use the `std` or `percentiles` arguments when creating the agent reporter. https://helipad.dev/functions/barchart/addbar/"""
		if not isinstance(color, Color): color = Color(color)
		bar = Item(reporter=reporter, label=label, color=color)

		#Add subsidiary series (e.g. percentile bars)
		bar.err = []
		if reporter in self.viz.model.data.reporters and self.viz.model.data.reporters[reporter].children:
			for p in self.viz.model.data.reporters[reporter].children:
				if '-unsmooth' in p: continue
				bar.err.append(p)

		if position is None or position>=len(self.bars): self.bars.append(bar)
		else: self.bars.insert(position-1, bar)

	def launch(self, axes):
		super().launch(axes)
		axes.spines['top'].set_visible(False)
		axes.spines['right'].set_visible(False)

		cfunc, eax = (axes.barh, 'xerr') if self.horizontal else (axes.bar, 'yerr')
		kwa = {eax: [0 for bar in self.bars]} #Make sure our error bars go the right way
		rects = cfunc(range(len(self.bars)), [0 for i in self.bars], color=[bar.color.hex for bar in self.bars], **kwa)
		errors = rects.errorbar.lines[2][0].properties()['paths'] #Hope MPL's API doesn't ever move this…!
		for bar, rect, err in zip(self.bars, rects, errors):
			bar.element = rect
			bar.errPath = err.vertices
			bar.errHist = []
			bar.data = []

		#Set bar names
		cstlfunc, cstfunc = (axes.set_yticklabels, axes.set_yticks) if self.horizontal else (axes.set_xticklabels, axes.set_xticks)
		cstfunc(range(len(self.bars)))
		cstlfunc([bar.label for bar in self.bars])
		axes.margins(x=0)

		if self.logscale:
			if self.horizontal:
				axes.set_xscale('log')
				axes.set_xlim(1/2, 2, auto=True)
			else:
				axes.set_yscale('log')
				axes.set_ylim(1/2, 2, auto=True)

	def update(self, data: dict, t: int):
		getlim, setlim = (self.axes.get_xlim, self.axes.set_xlim) if self.horizontal else (self.axes.get_ylim, self.axes.set_ylim)
		lims = list(getlim())
		for b in self.bars:
			b.data.append(data[b.reporter])
			if data[b.reporter] < lims[0]: lims[0] = data[b.reporter]
			if data[b.reporter] > lims[1]: lims[1] = data[b.reporter]

			if b.err:
				errs = [data[e] for e in b.err]
				errs.sort()
				for i,cap in enumerate(b.errPath): #Should only be 2 of these, but errs could be any length
					if errs[i] < lims[0]: lims[0] = errs[i]
					if errs[i] > lims[1]: lims[1] = errs[i]
				b.errHist.append(errs)

		setlim(lims)
		self.axes.autoscale_view(tight=False)

	def draw(self, t: int=None, forceUpdate: bool=False):
		if t is None: t=self.viz.scrubval
		i = int(t/self.viz.model.param('refresh'))-1
		for b in self.bars:
			setbar = b.element.set_width if self.horizontal else b.element.set_height
			setbar(b.data[i])

			if b.err: #Update error bars
				for j,cap in enumerate(b.errPath):
					if len(b.errHist[i]) >= j+1: cap[0 if self.horizontal else 1] = b.errHist[i][j]
		super().draw(t, forceUpdate)

class AgentsPlot(ChartPlot):
	"""A plot visualizer used for displaying individual agents in various layouts. Agents can be plotted in a network structure, a spatial layout (or both at once), or a scatterplot comparing agents on two data dimensions. https://helipad.dev/functions/agentsplot/"""
	type = 'agents'
	def __init__(self, **kwargs):
		if 'prim' not in kwargs: kwargs['prim'] = None
		if 'kind' in kwargs:
			warnings.warn(ï('The kind= argument is deprecated and has been replaced with network=.')) #Deprecated in Helipad 1.6; remove in Helipad 1.8
			kwargs['network'] = kwargs['kind']
		if 'network' not in kwargs: kwargs['network'] = 'edge'
		if 'layout' not in kwargs: kwargs['layout'] = 'scatter' if 'scatter' in kwargs else 'spring'
		if 'scatter' not in kwargs: kwargs['scatter'] = None
		super().__init__(**kwargs)
		self.scatterLims = [[0,0],[0,0]]
		self.ndata = {}

		self.params = {
			'patchColormap': 'Blues',
			#'patchProperty': 'mapcolor', #Don't specify, so we have a default white
			'agentMarker': 'o',
			'agentSize': 30,
			'agentLabel': False,
			'labelSize': 10,
			'labelColor': 'k',
			'labelFamily': 'sans-serif',
			'labelWeight': 'normal',
			'labelAlpha': None,
			'labelAlign': 'center',
			'labelVerticalAlign': 'center',
			'lockLayout': False,
			'patchAspect': 1,
			'mapBg': 'black',
			'regLine': False,
			'regColor': 'red',
			'regWidth': 1
		}

	def launch(self, axes):
		import networkx as nx, networkx.drawing.layout as lay
		self.layouts = {l: getattr(lay, l+'_layout') for l in ['spring', 'circular', 'kamada_kawai', 'random', 'shell', 'spectral', 'spiral']}

		def spatial_layout(G):
			if self.projection == 'polar': #Calculate the right angle from the x coordinate without modifying the original tuples
				return {i: (2*pi-(2*pi/self.viz.model.patches.dim[0] * data['position'][0])+1/2*pi, data['position'][1]) for i, data in G.nodes.items()}
			else: return {i: data['position'] for i,data in G.nodes.items()}

		def scatter_layout(G):
			self.axes.set_xlabel(self.scatter[0])
			self.axes.set_ylabel(self.scatter[1])
			self.axes.spines['top'].set_visible(False)
			self.axes.spines['right'].set_visible(False)
			data = {i: [data[self.scatter[0]], data[self.scatter[1]]] for i,data in G.nodes.items()}

			#Update limits if necessary
			xs, ys = array([d[0] for d in data.values()]), array([d[1] for d in data.values()])
			self.scatterLims[0][0] = min(self.scatterLims[0][0], xs.min())
			self.scatterLims[0][1] = max(self.scatterLims[0][1], xs.max())
			self.scatterLims[1][0] = min(self.scatterLims[1][0], ys.min())
			self.scatterLims[1][1] = max(self.scatterLims[1][1], ys.max())
			self.axes.set_xlim(*self.scatterLims[0])
			self.axes.set_ylim(*self.scatterLims[1])

			#Add regression line if applicable
			if self.params['regLine']:
				b0, b1 = linalg.lstsq(vstack([xs, ones(len(xs))]).T, ys, rcond=None)[0]
				self.axes.add_line(Line2D(
					[self.scatterLims[0][0], self.scatterLims[0][1]], [b0*self.scatterLims[0][0]+b1, b0*self.scatterLims[0][1]+b1],
					linestyle = '--' if self.params['regLine'] is True else self.params['regLine'],
					color = self.params['regColor'],
					linewidth = self.params['regWidth']
			    ))

			return data
		self.layouts['spatial'] = spatial_layout
		self.layouts['scatter'] = scatter_layout

		self.nx = nx
		self.components = {}

		super().launch(axes)

	def MPLEvent(self, event):
		if event.name=='key_press_event' and event.key=='l': self.rotateLayout()

		elif event.name=='pick_event':
			pk = list(self.pos.keys())
			agents = [self.viz.model.agents[pk[i]] for i in event.ind]
			self.viz.model.doHooks('agentClick', [agents, self, self.viz.scrubval])

		elif event.name=='button_press_event' and self.layout=='spatial':
			if self.projection=='polar':
				x = 2*pi-event.xdata+1/2*pi
				if x > 2*pi: x-=2*pi
				x,y = x/(2*pi/self.viz.model.patches.dim[0]), event.ydata
			else:
				x,y = event.xdata, event.ydata
			if x > self.viz.model.patches.boundaries[0][1] or y > self.viz.model.patches.boundaries[1][1]: return
			self.viz.model.doHooks('patchClick', [self.viz.model.patches.at(x,y), self, self.viz.scrubval])

	def update(self, data: dict, t: int):
		G = self.viz.model.agents.network(self.network, self.prim, excludePatches=self.prim!='patch')

		#Capture data for label, size, and scatterplot position
		capture = {}
		if self.params['agentLabel'] and self.params['agentLabel'] is not True: capture['label'] = self.params['agentLabel']
		if self.params['agentSize'] and type(self.params['agentSize']) not in [int, float]: capture['size'] = self.params['agentSize']
		if self.scatter:
			for v in self.scatter: capture[v] = v
		agents = {a.id:a for a in (self.viz.model.agents[self.prim] if self.prim else self.viz.model.agents.all)}
		for k,v in capture.items():
			if 'good:' in v:
				for n in G.nodes: G.nodes[n][k] = agents[n].stocks[v.split(':')[1]]
			else:
				for n in G.nodes: G.nodes[n][k] = getattr(agents[n],v)
		self.ndata[t] = G

		#Save spatial data even if we're on a different layout
		if self.viz.model.patches:
			lst = []
			for p in self.viz.model.agents['patch']:
				p.colorData[t] = self.getPatchParamValue(p, None if t==self.viz.model.t else t)
				lst.append(p.colorData[t])

			#Renormalize color scale
			nmin, nmax = min(lst), max(lst)
			self.normal = plt.cm.colors.Normalize(nmin if not hasattr(self,'normal') or nmin<self.normal.vmin else self.normal.vmin, nmax if not hasattr(self,'normal') or nmax>self.normal.vmax else self.normal.vmax)

	def draw(self, t: int=None, forceUpdate: bool=False):
		if t is None: t=self.viz.scrubval
		self.axes.clear()
		if self.layout not in ['spatial', 'scatter'] or self.projection=='polar': self.axes.axis('off')
		self.axes.set_title(self.label, fontdict={'fontsize':10})
		if self.layout != 'spatial' or self.viz.model.patches.geometry != 'geo':
			self.axes.set_facecolor('white')
			self.axes.set_aspect('auto')
		self.pos = self.layouts[self.layout](self.ndata[t])

		if self.layout == 'spatial':
			cmap = matplotlib.colormaps[self.params['patchColormap']]
			if self.viz.model.patches.geometry == 'geo': pd = array([self.getPatchParamValue(p,t) for p in self.viz.model.patches])
			else:
				pd = array([[self.getPatchParamValue(p,t) for p in col] for col in self.viz.model.patches]).T #Transpose because numpy is indexed col, row
			if self.projection=='polar':
				self.axes.set_aspect('equal')
				self.axes.set_ylim(0, self.viz.model.patches.dim[1])
				norm = (pd - int(self.normal.vmin))/(rng if (rng:=(self.normal.vmax-int(self.normal.vmin))) else 1)
				space = linspace(0.0, 1.0, 100)
				rgb = cmap(space)[newaxis, :, :3][0]

				x = self.viz.model.patches.dim[0]
				for r in range(len(norm)):
					for ti in range(len(norm[0])):
						#Define the range going counterclockwise. The 1/4 is to make r=0 point north rather than east.
						#Last argument is the resolution of the curve
						theta = arange(2*pi*((1-1/x) * ti - 1/x + 1/4), 2*pi*((1-1/x) * ti + 1/4)+.04, .04)
						color = self.params['mapBg'] if isnan(norm[r, ti]) else rgb[int(norm[r,ti]*(len(space)-1))]
						self.axes.fill_between(theta, full_like(theta, r), full_like(theta, r+1), color=color)
			else:
				if self.viz.model.patches.geometry == 'rect':
					self.axes.spines['top'].set_visible(True)
					self.axes.spines['right'].set_visible(True)
					cmap.set_bad(self.params['mapBg'])
					self.components['patches'] = self.axes.imshow(pd, norm=self.normal, cmap=cmap, aspect=self.params['patchAspect'])
					# self.patchmap.set_norm(self.normal)
					# self.components['patches'].set_data(pd)
				elif self.viz.model.patches.geometry == 'geo':
					self.axes.set_aspect('equal')
					self.axes.set_xlim(*self.viz.model.patches.boundaries[0])
					self.axes.set_ylim(*self.viz.model.patches.boundaries[1])
					self.axes.set_facecolor(self.params['mapBg'])
					norm = (pd - int(self.normal.vmin))/(rng if (rng:=(self.normal.vmax-int(self.normal.vmin))) else 1)
					space = linspace(0.0, 1.0, 100)
					rgb = cmap(space)[newaxis, :, :3][0]
					for i,p in enumerate(self.viz.model.patches):
						color = self.params['mapBg'] if isnan(norm[i]) else rgb[int(norm[i]*(len(space)-1))]
						self.axes.add_patch(Polygon(array(p.polygon.exterior.xy).T, color=color))

		#Draw nodes, edges, and labels separately so we can split out the directed and undirected edges
		sizes = self.params['agentSize']*10 if type(self.params['agentSize']) in [int, float] else [n[1]['size']*10 for n in self.ndata[t].nodes(data=True)]
		self.components['nodes'] = self.nx.draw_networkx_nodes(self.ndata[t], self.pos, ax=self.axes, node_color=[self.viz.model.agents[n[1]['primitive']].breeds[n[1]['breed']].color.hex for n in self.ndata[t].nodes(data=True)], node_size=sizes, node_shape=self.params['agentMarker'])
		e_directed = [e for e in self.ndata[t].edges.data() if e[2]['directed']]
		e_undirected = [e for e in self.ndata[t].edges.data() if not e[2]['directed']]
		self.components['edges_d'] = self.nx.draw_networkx_edges(self.ndata[t], self.pos, ax=self.axes, edgelist=e_directed, width=[e[2]['weight'] for e in e_directed])
		self.components['edges_u'] = self.nx.draw_networkx_edges(self.ndata[t], self.pos, ax=self.axes, edgelist=e_undirected, width=[e[2]['weight'] for e in e_undirected], arrows=False)
		if self.params['agentLabel']:
			lab = None if self.params['agentLabel'] is True else {n:self.ndata[t].nodes[n]['label'] for n in self.ndata[t].nodes}
			self.components['labels'] = self.nx.draw_networkx_labels(self.ndata[t], self.pos, ax=self.axes, labels=lab, font_size=self.params['labelSize'], font_color=self.params['labelColor'], font_family=self.params['labelFamily'], font_weight=self.params['labelWeight'], alpha=self.params['labelAlpha'], horizontalalignment=self.params['labelAlign'], verticalalignment=self.params['labelVerticalAlign'])
		if self.layout=='scatter': self.axes.tick_params('both', left=True, bottom=True, labelleft=True, labelbottom=True)

		self.components['nodes'].set_picker(True)	#Listen for mouse events on nodes
		self.components['nodes'].set_pickradius(5)	#Set margin of valid events in pixels

		super().draw(t, forceUpdate)

	def rotateLayout(self):
		"""Rotate the layout among the several NetworkX layout classes, plus spatial and scatterplot layouts. This method is called when pressing 'l' with the mouse over the plot. https://helipad.dev/functions/agentsplot/rotatelayout/"""
		self.axes.set_yscale('linear') #Override default logscale keypress
		if self.params['lockLayout']: return

		#Select the next layout in the list
		layouts = ['spatial'] if self.viz.model.patches else []
		if self.scatter: layouts.append('scatter')
		if self.network in self.viz.model.agents.edges or not layouts: layouts += ['spring', 'circular', 'kamada_kawai', 'random', 'shell', 'spectral', 'spiral']
		li = layouts.index(self.layout)+1
		while li>=len(layouts): li -= len(layouts)
		self.layout = layouts[li]

		#Replace the axes object if we need to switch projections
		if self.projection == 'polar' or (self.layout=='spatial' and self.viz.model.patches and self.viz.model.patches.geometry=='polar'):
			self.projection = None if self.projection=='polar' else 'polar'
			self.viz.fig.delaxes(self.axes)
			super().launch(self.viz.fig.add_subplot(*self.subplot_position, projection=self.projection))

		#kamada_kawai requires scipy; fail silently and continue if we don't have it
		try: self.draw(self.viz.scrubval)
		except: self.rotateLayout()

	def getPatchParamValue(self, patch, t: int=None):
		"""Gather historical patch data for use in color generation. https://helipad.dev/functions/agentsplot/getpatchparamvalue/"""
		if t is not None: return patch.colorData[t]
		if patch.dead: return float('nan')
		elif 'patchProperty' not in self.params: return 0
		elif 'good:' in self.params['patchProperty']: return patch.stocks[self.params['patchProperty'].split(':')[1]]
		else: return getattr(patch, self.params['patchProperty'])

	def config(self, param: str, val=None):
		"""Sets options for the spatial plot. One argument will return the value; two arguments sets the parameter. See https://helipad.dev/functions/agentsplot/config/ for valid option names."""
		if isinstance(param, dict):
			for k,v in param.items(): self.config(k,v)
		elif val is None: return self.params[param]
		else: self.params[param] = val

#======================
# SUB-PLOT OBJECTS
# Mostly sub-plot data will use an Item object, but some need a little more
#======================

class Series(Item):
	"""Stores data on series, which link a reporter to a plot. https://helipad.dev/functions/series/"""
	@property
	def visible(self) -> bool:
		"""Whether the series is to be drawn when the plot window launches, or - if it has launched - whether the series is currently visible. https://helipad.dev/functions/series/#visible"""
		if not hasattr(self, 'line'): return self._visible
		elif hasattr(self, 'poly'): return True #If it's a stackplot
		else: return self.line.get_visible()

	@visible.setter
	def visible(self, val: bool):
		if not hasattr(self, 'line'): #If we haven't drawn it yet
			self._visible = val
			return
		elif hasattr(self, 'poly'): return	#Ignore if it's a stackplot

		self.line.set_visible(val)
		for s in self.subseries: s.line.set_visible(val) #Toggle subseries (e.g. percentile bars)
		self.legline.set_alpha(1.0 if val else 0.2)
		self.legtext.set_alpha(1.0 if val else 0.2)

		## Won't work because autoscale_view also includes hidden lines
		## Will have to actually remove and reinstate the line for this to work
		# for g in self.activePlots:
		# 	if c1.series in g.get_lines():
		# 		g.relim()
		# 		g.autoscale_view(tight=True)

		self.line.figure.canvas.draw_idle()

	def toggle(self):
		"""Toggles the visibility of the series. https://helipad.dev/functions/series/toggle/"""
		self.visible = not self.visible

#======================
# HELPER FUNCTIONS
#======================

def keepEvery(lst: list, n: int):
	"""Reduce the resolution of a list."""
	i,l = (1, [])
	for k in lst:
		if i%n==0: l.append(k)
		i+=1
	# if len(lst)%n != 0: print('Original:',len(lst),'New:',len(l),'Remainder:',len(lst)%n)
	return l


"""
Classes for collecting and exporting data from model runs. This module should not be imported directly; interface with the `model.data` container object instead. See https://helipad.dev/functions/data/
"""

import os.path
import pandas, numpy as np
from helipad.helpers import Item, ï

#Don't try to subclass Pandas.dataframe; substantially slower and doesn't scale
class Data:
	"""Interface for collecting, storing, and accessing data generated during model runs. The object is subscriptable: `data[key]` will return the column of the data labelled `key`. Stored in `model.data`. https://helipad.dev/functions/data/"""
	def __init__(self, model):
		self.reporters = {}
		self.model = model

	def __getitem__(self, index):
		r = self.columns[index]
		return r.data if index==r.name else r.children[index][1]

	def __contains__(self, index): return index in self.columns
	def __repr__(self): return f'<{self.__class__.__name__}: {len(self.reporters)} reporters>'

	def addReporter(self, key: str, func, **kwargs):
		"""Register a column in the data to be collected each period. `func` can either be a function that takes the model object as its only argument and returns a value, or a string - either `model` or the name of a primitive. Kwargs are passed to the reporter class. https://helipad.dev/functions/data/addreporter/"""
		func, children = func if isinstance(func, tuple) else (func, {})

		if not callable(func): raise TypeError(ï('Second argument of addReporter must be callable.'))
		self.reporters[key] = Reporter(name=key, func=func, children=children, **kwargs)
		return self.reporters[key]

	def removeReporter(self, key: str):
		"""Remove a column and its subsidiaries (e.g. percentiles) from the data collection, stop its reporter, and remove any associated series. https://helipad.dev/functions/data/removereporter/"""
		if self.model.hasModel:
			raise RuntimeError(ï('removeReporter cannot be called while a model is active.'))
		self.model.doHooks('removeReporter', [self, key])
		del self.reporters[key]

	def collect(self, model):
		"""Iterate over all the registered reporters and collect model data each period. This function is called from `model.step()` and should not be called from user code. https://helipad.dev/functions/data/collect/"""
		model.doHooks('dataCollect', [self, model.t])
		for v in self.reporters.values(): v.collect(model)

	def reset(self):
		"""Clear all model data. Generally used to clean up between model runs. https://helipad.dev/functions/data/reset/"""
		for v in self.reporters.values(): v.clear()

	@property
	def all(self) -> dict:
		"""A dict of all model data, with keys corresponding to registered reporters. https://helipad.dev/functions/data/#all"""
		data = {}
		for k,r in self.reporters.items():
			data[k] = r.data
			for s,d in r.children.items(): data[s] = d[1]
		return data

	@property
	def columns(self) -> dict:
		"""A `dict` of data columns and the associated Reporter objects. This property does not correspond to `Data.reporters.keys()`, because multiple columns may be associated with the same reporter (e.g. with percentile bars, smoothing, or ± standard deviations). https://helipad.dev/functions/data/#columns"""
		cols = {}
		for k,r in self.reporters.items():
			cols[k] = r
			for s in r.children: cols[s] = r
		return cols

	@property
	def dataframe(self):
		"""A `Pandas` dataframe with the model run data. https://helipad.dev/functions/data/#dataframe"""
		return pandas.DataFrame(self.all)

	#
	# REPORTERS
	# These functions return reporters; they take some arguments and return a function that takes 1 argument: model.
	#

	def modelReporter(self, key):
		"""Generate a reporter function that takes the `model` object and returns the named attribute of the model. https://helipad.dev/functions/data/modelreporter/"""
		def reporter(model): return getattr(model, key)
		return reporter

	# NOTE: Batching data collection (looping over agents and then variables, instead of – as now – looping over
	# variables and then agents) did not result in any speed gains; in fact a marginal (0.65%) speed reduction
	def agentReporter(self, key: str, prim=None, breed=None, good=None, stat: str='mean', **kwargs):
		"""Generate a reporter function that takes the `model` object and returns a summary statistic (`'mean'`, `'sum'`, `'gmean'` (for geometric mean), `'std'` (for standard deviation), or `'percentile-nn'`, where `nn` is a number from 0-100.) over all the values of an agent property. https://helipad.dev/functions/data/agentreporter/"""
		if prim is None: prim = next(iter(self.model.agents))
		# if breed and isinstance(breed, bool): return [self.agentReporter(key+'-'+br, prim, br, good, stat, **kwargs) for br in self.model.agents[prim].breeds]
		if 'percentiles' in kwargs:
			subplots = {('' if not breed else breed)+key+'-'+str(p)+'-pctile':self.agentReporter(key, prim, breed=breed, good=good, stat='percentile-'+str(p)) for p in kwargs['percentiles']}
		elif 'std' in kwargs:
			subplots = {('' if not breed else breed)+key+'+'+str(kwargs['std'])+'std': self.agentReporter(key, prim, breed=breed, good=good, stat='mstd-p-'+str(kwargs['std'])), key+'-'+str(kwargs['std'])+'std': self.agentReporter(key, prim, breed=breed, good=good, stat='mstd-m-'+str(kwargs['std']))}
		else: subplots = None

		def reporter(model):
			#Construct list of values
			array = [getattr(a, key) for a in (model.agents.all if prim=='all' else model.agents[prim]) if breed is None or breed==a.breed]
			if good is not None: array = [v[good] for v in array]
			u = [v for v in array if v is not None]

			if not u: return 0
			elif stat=='sum':	return sum(u)
			elif stat=='mean':	return np.mean(u)
			elif stat=='gmean':	return np.exp(np.log(u).sum()/len(u))
			elif stat=='std':	return np.std(u)
			elif stat=='max':	return max(u)
			elif stat=='min':	return min(u)
			elif 'percentile-' in stat:
				pctile = int(stat.split('-')[1])
				if pctile==100: return max(u)
				elif pctile==0: return min(u)
				else:
					u.sort()
					idx = round(len(u)*pctile/100)
					return u[idx] if len(u)>=idx+1 else u[0]
			elif 'mstd-' in stat: #Don't use directly; use the std kwarg
				s, op, coef = stat.split('-')
				coef = float(coef)
				if op=='p': return np.mean(u) + coef * np.std(u)
				else: return np.mean(u) - coef * np.std(u)
			else: raise ValueError(ï('Invalid statistic {}.').format(stat))
		return (reporter, subplots) if subplots is not None else reporter

	#
	# OTHER FUNCTIONS
	#

	#Slices the last n records for the key registered as a reporter
	#Returns a value if n=1, otherwise a list
	def getLast(self, key: str, n: int=1):
		"""Return the latest recorded value or values from the model's data. https://helipad.dev/functions/data/getlast/"""
		if isinstance(key, str):
			if len(self[key])==0: return 0
			data = self[key][-n:]
			return data if n>1 else data[0]
		elif isinstance(key, int):
			return {k: v[-key:] for k,v in self.all.items()}
		else: raise TypeError(ï('First argument of Data.getLast() must be either a key name or an int.'))

	def saveCSV(self, filename: str='data'):
		"""Outputs the model's data to a CSV file in the same directory as the running program. https://helipad.dev/functions/data/savecsv/"""
		file = filename + '.csv'
		i=0
		while os.path.isfile(file): #Avoid filename collisions
			i += 1
			file = filename+'-'+str(i)+'.csv'

		df = self.dataframe
		hook = self.model.doHooks('saveCSV', [df, self.model]) #can't do `or None` since "The truth value of a DataFrame is ambiguous"
		if hook is not None: df = hook
		df.to_csv(file)

class Reporter(Item):
	"""An interface defining a single column of data to be collected during a model run. https://helipad.dev/functions/reporter/"""
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.data = []
		self.children = {k:(fn, []) for k, fn in self.children.items()}
		if 'smooth' not in kwargs: self.smooth = False
		else: self.children[self.name+'-unsmooth'] = (None, [])

	def collect(self, model):
		"""Run the reporter function and its children, and append the results to the data. https://helipad.dev/functions/reporter/collect/"""
		for s in self.children.values():
			if callable(s[0]): s[1].append(s[0](model))

		if not self.smooth: self.data.append(self.func(model))
		else:
			us = self.children[self.name+'-unsmooth'][1]
			us.append(self.func(model))
			#					  Old data 					   New data point
			self.data.append(self.smooth*self.data[-1] + (1-self.smooth)*us[-1] if model.t>1 else us[-1])

	def clear(self):
		"""Empty the reporter's collected data. https://helipad.dev/functions/reporter/clear/"""
		self.data.clear()
		for c in self.children.values(): c[1].clear()


# A decentralized price discovery model with random matching

#===============
# SETUP
#===============

from helipad import Helipad
from helipad.utility import CobbDouglas
from math import exp, floor
import random

heli = Helipad()
heli.name = 'Price Discover'
heli.agents.order = 'match'

heli.params.add('ratio', 'Log Endowment Ratio', 'slider', dflt=0, opts={'low': -3, 'high': 3, 'step': 0.5}, runtime=False)
heli.params['num_agent'].opts['step'] = 2 #Make sure we don't get stray agents
heli.params['num_agent'].opts['low'] = 2 #Make sure we don't get stray agents

heli.goods.add('shmoo','#11CC00', (1, 1000))
heli.goods.add('soma', '#CC0000', lambda breed: (1, floor(exp(heli.param('ratio'))*1000)))

#===============
# BEHAVIOR
#===============

@heli.hook
def agentInit(agent, model):
	agent.utility = CobbDouglas(['shmoo', 'soma'])

@heli.hook
def match(agents, primitive, model, stage):
	myEndowU = agents[0].utility.calculate(agents[0].stocks)
	theirEndowU = agents[1].utility.calculate(agents[1].stocks)

	#Get the endpoints of the contract curve
	#Contract curve isn't linear unless the CD exponents are both 0.5. If not, *way* more complicated
	cc1Soma = myEndowU * (sum(a.stocks['soma'] for a in agents)/sum(a.stocks['shmoo'] for a in agents)) ** 0.5
	cc2Soma = sum(a.stocks['soma'] for a in agents) - theirEndowU  * (sum(a.stocks['soma'] for a in agents)/sum(a.stocks['shmoo'] for a in agents)) ** 0.5
	cc1Shmoo = sum(a.stocks['shmoo'] for a in agents)/sum(a.stocks['soma'] for a in agents) * cc1Soma
	cc2Shmoo = sum(a.stocks['shmoo'] for a in agents)/sum(a.stocks['soma'] for a in agents) * cc2Soma

	#Calculate demand: choose a random point on the contract curve
	r = random.random()
	somaDemand = r*cc1Soma + (1-r)*cc2Soma - agents[0].stocks['soma']
	shmooDemand = r*cc1Shmoo + (1-r)*cc2Shmoo - agents[0].stocks['shmoo']

	#Do the trades
	if abs(somaDemand) > 0.1 and abs(shmooDemand) > 0.1:
		agents[0].trade(agents[1], 'soma', -somaDemand, 'shmoo', shmooDemand)
		agents[0].lastPrice = -somaDemand/shmooDemand
		agents[1].lastPrice = -somaDemand/shmooDemand
	else:
		agents[0].lastPrice = None
		agents[1].lastPrice = None

	#Record data
	agents[0].utils = agents[0].utility.calculate(agents[0].stocks)
	agents[1].utils = agents[1].utility.calculate(agents[1].stocks)

#Stop the model when we're basically equilibrated
@heli.event
def stopCondition(model):
	return model.t > 1 and model.data.getLast('demand-shmoo') < 20 and model.data.getLast('demand-soma') < 20
heli.param('stopafter', 'stopCondition')

#===============
# DATA AND VISUALIZATION
#===============

from helipad.visualize import TimeSeries
viz = heli.useVisual(TimeSeries)

heli.data.addReporter('ssprice', heli.data.agentReporter('lastPrice', 'agent', stat='gmean', percentiles=[0,100]))
pricePlot = viz.addPlot('price', 'Price', logscale=True, selected=True)
pricePlot.addSeries('ssprice', 'Soma/Shmoo Price', '#119900')

for p in ['demand', 'utility']: viz[p].active(True)
heli.param('refresh', 1)

heli.launchCpanel()


# A model of the relative price effects of monetary shocks via helicopter drop vs. by open market operations.
# Includes a banking system to distribute OMOs. Requires the basic Helicopter.py in the same folder.
# Download the paper at https://ssrn.com/abstract=2545488

from numpy import mean, array #, std
from Helicopter import * #Building on the basic Helicopter model
heli = setup()

#===============
# BANK CLASS
#===============

# baseAgent.overdraft = 'continue-warn'
class Bank(baseAgent):
	def __init__(self, breed, id, model):
		super().__init__(breed, id, model)

		self.i = .1				#Per-period interest rate
		self.targetRR = 0.25
		self.inflation = 0
		self.accounts = {}		#Liabilities
		self.credit = {}		#Assets

		self.dif = 0			#How much credit was rationed
		self.defaultTotal = 0
		self.pLast = 50 		#Initial price level, equal to average of initial prices

	def account(self, customer):
		return self.accounts[customer.id] if customer.id in self.accounts else 0

	def setupAccount(self, customer):
		if customer.id in self.accounts: return False		#If you already have an account
		self.accounts[customer.id] = 0						#Liabilities
		self.credit[customer.id] = Loan(customer, self)		#Assets

	#Assets and liabilities should return the same thing
	#Any difference gets disbursed as interest on deposits
	@property
	def assets(self):
		return self.stocks[self.model.goods.money] + sum(l.owe for l in self.credit.values()) #Reserves

	@property
	def liabilities(self):
		return sum(list(self.accounts.values())) #Values returns a dict_values object, not a list. So wrap it in list()

	@property
	def loans(self):
		return self.assets - self.stocks[self.model.goods.money]

	@property
	def reserveRatio(self):
		l = self.liabilities
		if l == 0: return 1
		else: return self.stocks[self.model.goods.money] / l

	@property
	def realInterest(self): return self.i - self.inflation

	#amt<0 to withdraw
	def deposit(self, customer, amt):
		amt = customer.pay(self, amt)
		self.accounts[customer.id] += amt	#Credit account
		return amt

	def transfer(self, customer, recipient, amt):
		if self.accounts[customer.id] < amt: amt = self.accounts[customer.id]
		self.accounts[customer.id] -= amt
		self.accounts[recipient.id] += amt
		return amt

	def borrow(self, customer, amt):
		if amt < 0.01: return 0 #Skip blanks and float errors
		l = self.credit[customer.id]

		#Refinance anything with a higher interest rate
		for n,loan in enumerate(l.loans):
			if loan['i'] >= self.i:
				amt += loan['amount']
				del l.loans[n]

		#Increase assets
		l.loans.append({
			'amount': amt,
			'i': self.i
		})

		self.accounts[customer.id] += amt	#Increase liabilities
		return amt							#How much you actually borrowed

	#Returns the amount you actually pay – the lesser of amt or your outstanding balance
	def amortize(self, customer, amt):
		if amt < 0.001: return 0			#Skip blanks and float errors
		l = self.credit[customer.id]	#Your loan object
		l.amortizeAmt += amt				#Count it toward minimum repayment
		leftover = amt

		#Reduce assets; amortize in the order borrowed
		while leftover > 0 and len(l.loans) > 0:
			if leftover >= l.loans[0]['amount']:
				leftover -= l.loans[0]['amount']
				del l.loans[0]
			else:
				l.loans[0]['amount'] -= leftover
				leftover = 0

		self.accounts[customer.id] -= (amt - leftover)	#Reduce liabilities
		return amt - leftover							#How much you amortized

	def step(self, stage):
		for l in self.credit.values(): l.step()

		#Pay interest on deposits
		lia = self.liabilities
		profit = self.assets - lia
		if profit > self.model.param('num_agent'):
			print('Disbursing profit of $',profit)
			for id, a in self.accounts.items():
				self.accounts[id] += profit/lia * a

		#Calculate inflation as the unweighted average price change over all goods
		if self.model.t >= 2:
			inflation = self.model.cb.P/self.pLast - 1
			self.pLast = self.model.cb.P	#Remember the price from this period before altering it for the next period
			self.inflation = (19 * self.inflation + inflation) / 20		#Decaying average

		#Set interest rate and/or minimum repayment schedule
		#Count potential borrowing in the interest rate adjustment
		targeti = self.i * self.targetRR / (self.reserveRatio)

		#Adjust in proportion to the rate of reserve change
		#Positive deltaReserves indicates falling reserves; negative deltaReserves rising inventory
		if self.model.t > 2:
			deltaReserves = (self.lastReserves - self.stocks[self.model.goods.money])/self.model.cb.P
			targeti *= (1 + deltaReserves/(20 ** self.model.param('pSmooth')))
		self.i = (self.i * 24 + targeti)/25										#Interest rate stickiness

		self.lastReserves = self.stocks[self.model.goods.money]

		#Upper and lower interest rate bounds
		self.i = min(self.i, 1+self.inflation)		#interest rate cap at 100%
		self.i = max(self.i, self.inflation+0.005)	#no negative real rates
		self.i = max(self.i, 0.005)					#no negative nominal rates

class Loan():
	def __init__(self, customer, bank):
		self.customer = customer
		self.bank = bank
		self.loans = []
		self.amortizeAmt = 0

	@property
	def owe(self): return sum([l['amount'] for l in self.loans])

	def step(self):
		#Charge the minimum repayment if the agent hasn't already amortized more than that amount
		minRepay = 0
		for l in self.loans:
			iLoan = l['amount'] * l['i']
			minRepay += iLoan				#You have to pay at least the interest each period
			l['amount'] += iLoan			#Roll over the remainder at the original interest rate

		#If they haven't paid the minimum this period, charge it
		amtz = minRepay - self.amortizeAmt
		defaulted = False
		if amtz > 0:
			if amtz > self.bank.accounts[self.customer.id]:	#Can't charge them more than they have in the bank
				defaulted = True
				amtz = self.bank.accounts[self.customer.id]
				# print(self.bank.model.t, ': Agent', self.customer.id, 'defaulted $', self.owe - amtz)
			self.bank.amortize(self.customer, amtz)
			if defaulted:
				for n, l in enumerate(self.loans):
					self.loans[n]['amount'] /= 2
					self.bank.defaultTotal += l['amount']/2
					##Cap defaults at the loan amount. Otherwise if i>1, defaulting results in negative debt
					# if l['i'] >= 1:
					# 	self.bank.defaultTotal += l['amount']
					# 	del self.loans[n]
					# else:
					# 	l['amount'] -= l['amount'] * l['i']
					# 	self.bank.defaultTotal += l['amount'] * l['i']

		self.amortizeAmt = 0

#===============
# CONFIGURATION
#===============

heli.agents.addPrimitive('bank', Bank, dflt=1, priority=1, hidden=True)
heli.name = 'Helicopter/OMO'

#Disable the irrelevant checkboxes if the banking model isn't selected
#Callback for the dist parameter
@heli.hook('terminate') #Reset the disabled checkmarks when terminating a model
def bankChecks(model, val=None):
	nobank = model.param('dist')!='omo'
	model.param('num_bank', 0 if nobank else 1)
	for i in ['debt', 'rr', 'i']:
		model.params['plots'].element.disabled(i, nobank)
	for e in model.params['liqPref'].elements.values():
		e.config(state='disabled' if nobank else 'normal')

heli.hooks.add('CpanelPostInit', lambda cpanel: bankChecks(cpanel.model))	#Set the disabled checkmarks on initialization

# UPDATE CALLBACKS

def lpUpdater(model, var, breed, val):
	if model.hasModel:
		for a in model.agents['agent']:
			if a.breed == breed: a.liqPref = val

heli.params['dist'].opts = {
	'prop': 'Helicopter/Proportional',
	'lump': 'Helicopter/Lump Sum',
	'omo': 'Open Market Operation'
}
#Since the param callback takes different parameters than the GUI callback
heli.params['dist'].callback = lambda model, var, val: bankChecks(model, val)
heli.param('dist','omo')

heli.params.add('liqPref', 'Demand for Liquidity', 'slider', per='breed', dflt={'hobbit': 0.1, 'dwarf': 0.3}, opts={'low':0, 'high': 1, 'step': 0.01}, prim='agent', callback=lpUpdater, desc='The proportion of the agent\'s balances he desires to keep in cash')

heli.visual.addPlot('debt', 'Debt', selected=False)
heli.visual.addPlot('rr', 'Reserve Ratio', selected=False)
heli.visual.addPlot('i', 'Interest Rate', selected=False)

#================
# AGENT BEHAVIOR
#================

#
# General
#

#Don't bother keeping track of the bank-specific variables unless the banking system is there
#Do this here rather than at the beginning so we can decide at runtime

def cashdemand(model):
	return model.data.agentReporter('stocks', good=model.goods.money, stat='sum', prim='agent')(model)/model.data.agentReporter('balance', stat='sum', prim='agent')(model)

@heli.hook
def modelPreSetup(model):
	if model.param('num_bank') > 0:
		model.data.addReporter('defaults', model.data.agentReporter('defaultTotal', 'bank'))
		model.data.addReporter('debt', model.data.agentReporter('loans', 'bank'))
		model.data.addReporter('reserveRatio', model.data.agentReporter('reserveRatio', 'bank'))
		model.data.addReporter('targetRR', model.data.agentReporter('targetRR', 'bank'))
		model.data.addReporter('i', model.data.agentReporter('i', 'bank'))
		model.data.addReporter('r', model.data.agentReporter('realInterest', 'bank'))
		model.data.addReporter('cashdemand', cashdemand)
		model.data.addReporter('inflation', model.data.agentReporter('inflation', 'bank'))
		model.data.addReporter('M2', lambda model: model.cb.M2)

		model.visual['money'].addSeries('defaults', 'Defaults', '#CC0000')
		model.visual['money'].addSeries('M2', 'Money Supply', '#000000')
		model.visual['debt'].addSeries('debt', 'Outstanding Debt', '#000000')
		model.visual['rr'].addSeries('targetRR', 'Target', '#777777')
		model.visual['rr'].addSeries('reserveRatio', 'Reserve Ratio', '#000000')
		model.visual['i'].addSeries('i', 'Nominal interest', '#000000')
		model.visual['i'].addSeries('r', 'Real interest', '#0000CC')
		model.visual['i'].addSeries('inflation', 'Inflation', '#CC0000')
		model.visual['i'].addSeries('cashdemand', '% cash held', '#00CC00')

@heli.hook
def modelPostSetup(model):
	if hasattr(model.agents['store'][0], 'bank'):
		model.agents['store'][0].pavg = 0
		model.agents['store'][0].projects = []
		model.agents['store'][0].defaults = 0

@heli.hook
def storeStep(agent, model, stage):
	#Intertemporal transactions
	if hasattr(agent, 'bank') and model.t > 0:
		#Stipulate some demand for credit, we can worry about microfoundations later
		agent.bank.amortize(agent, agent.bank.credit[agent.id].owe/1.5) #Pay back 2/3 of the outstanding balance each period
		agent.bank.borrow(agent, model.cb.ngdp * (1-agent.bank.i)) #Borrow some amount rising in sales volume and declining in the interest rate

#
# Agents
#

#Choose a bank if necessary
@heli.hook
def baseAgentInit(agent, model):
	if model.param('num_bank') > 0 and agent.primitive != 'bank':
		agent.bank = model.agents['bank'][0]
		agent.bank.setupAccount(agent)

@heli.hook
def agentInit(agent, model):
	if model.param('num_bank') > 0:
		agent.liqPref = model.param(('liqPref', agent.breed))

@heli.hook
def agentStep(agent, model, stage):
	#Deposit cash in the bank at the end of each period
	if hasattr(agent, 'bank'):
		tCash = agent.liqPref * (1/(1+agent.bank.i)) * agent.balance #Slightly interest-elastic
		agent.bank.deposit(agent, agent.stocks[agent.model.goods.money]-tCash)

#Use the bank if the bank exists
@heli.hook
def buy(agent, partner, good, q, p):
	if hasattr(agent, 'bank'):
		bal = agent.bank.account(agent)
		if p*q > bal:
			amount = bal
			leftover = (p*q - bal)/q
		else:
			amount = p*q
			leftover = 0
		agent.bank.transfer(agent, partner, amount)
		return (q, leftover)

#Use the bank if the bank exists
@heli.hook
def pay(agent, recipient, amount, model):
	if hasattr(agent, 'bank') and recipient.primitive != 'bank' and agent.primitive != 'bank':
		bal = agent.bank.account(agent)
		if amount > bal: #If there are not enough funds
			trans = bal
			amount -= bal
		else:
			trans = amount
			amount = 0
		agent.bank.transfer(agent, recipient, trans)
		return amount #Should be zero. Anything leftover gets paid in cash

@heli.hook
def checkBalance(agent, balance, model):
	if hasattr(agent, 'bank') and agent.primitive != 'bank':
		balance += agent.bank.account(agent)
		return balance

#
# Central Bank
#

def cbstep(self):
	#Record macroeconomic vars at the end of the last stage
	#Getting demand has it lagged one period…
	self.ngdp = sum(self.model.data.getLast('demand-'+good) * self.model.agents['store'][0].price[good] for good in self.model.goods.nonmonetary)
	if not self.ngdpAvg: self.ngdpAvg = self.ngdp
	self.ngdpAvg = (2 * self.ngdpAvg + self.ngdp) / 3

	#Set macroeconomic targets
	expand = 0
	if self.ngdpTarget: expand = self.ngdpTarget - self.ngdpAvg
	if self.model.param('num_bank') > 0: expand *= self.model.agents['bank'][0].reserveRatio
	if expand != 0: self.expand(expand)
CentralBank.step = cbstep

def expand(self, amount):

	#Deposit with each bank in proportion to their liabilities
	if 'bank' in self.model.agents and self.model.param('num_bank') > 0:
		self.stocks[self.model.goods.money] += amount
		self.model.agents['bank'][0].deposit(self, amount) #Budget constraint taken care of in .pay()

	elif self.model.param('dist') == 'lump':
		amt = amount/self.model.param('num_agent')
		for a in self.model.agents['agent']:
			a.stocks[self.model.goods.money] += amt
	else:
		M0 = self.M0
		for a in self.model.agents.all:
			a.stocks[self.model.goods.money] += a.stocks[self.model.goods.money]/M0 * amount
CentralBank.expand = expand

def M2(self):
	if 'bank' not in self.model.agents or self.model.param('num_bank') == 0: return self.M0
	return sum(a.balance for a in self.model.agents.all)
CentralBank.M2 = property(M2)

#Price level
#Average good prices at each store, then average all of those together weighted by the store's sale volume
#Figure out whether to break this out or not
def cbP(self):
	# denom = 0
	# numer = 0
	if not 'store' in self.model.agents: return None
	return mean(array(list(self.model.agents['store'][0].price.values())))
	# for s in self.model.agents['store']:
	# 	volume = sum(list(s.lastDemand.values()))
	# 	numer += mean(array(list(s.price.values()))) * volume
	# 	denom += volume
	#
	# if denom==0: return 1
	# else: return numer/denom
CentralBank.P = property(cbP)

heli.launchCpanel()


# An implementation of Conway's Game Of Life
# https://en.wikipedia.org/wiki/Conway%27s_Game_of_Life

from helipad import Helipad
from random import getrandbits

heli = Helipad()
heli.name = 'Game of Life'
heli.stages = 2
heli.param('refresh', 1)

heli.agents.removePrimitive('agent')
mapPlot = heli.spatial(dim=30, wrap=True, corners=True)
mapPlot.config('patchProperty', 'active')

def setdim(model, var, val): model.patches.dim = (int(val), int(val))
heli.params.add('dim', 'Dimension', 'slider', 30, opts={'low': 4, 'high': 30, 'step': 1}, runtime=False, callback=setdim)

@heli.hook
def patchInit(patch, model):
	patch.active = False

@heli.hook
def patchStep(patch, model, stage):
	if stage==1: #Calculate neighbors *before* dying or coming to life
		patch.activeNeighbors = len([n for n in patch.neighbors if n.active])
	elif stage==2:
		if patch.active and not 1 < patch.activeNeighbors < 4: patch.active = False
		elif not patch.active and patch.activeNeighbors == 3: patch.active = True

#Pause on launch to allow the user to toggle patches
@heli.hook
def modelStep(model, stage):
	if model.t==1 and stage==1: model.stop()

@heli.hook
def patchClick(patch, plot, t):
	patch.active = not patch.active
	plot.update(None, t)
	plot.draw(t, forceUpdate=True)

@heli.button
def Randomize(model):
	if not model.hasModel: return
	for p in model.agents['patch']:
		p.active = bool(getrandbits(1))
	if model.visual:
		model.visual['map'].update(None, model.t)
		model.visual['map'].draw(model.t, forceUpdate=True)

heli.launchCpanel()


"""
Functions and classes to initialize a network of Patch agents and add orientation and motion functions to other agents. This module should not be imported directly; use `model.spatial()` to set up a spatial model instead. See https://helipad.dev/functions/model/spatial/
"""

import warnings
from random import uniform
from math import sqrt, sin, cos, atan2, pi, copysign, floor
from abc import ABC, abstractmethod
from numbers import Number
from helipad.agent import Patch, baseAgent
from helipad.visualize import Charts
from helipad.helpers import ï, Item

#===============
# SETUP
# Create parameters, add functions, and so on
#===============

def spatialSetup(model, dim=10, corners=False, geometry: str='rect', offmap: bool=False, **kwargs):
	"""Set up functions, properties, and methods for a spatial model. https://helipad.dev/functions/model/spatial/"""

	#Initialize patch container and primitive
	model.agents.addPrimitive('patch', Patch, hidden=True, priority=-10)
	pClasses = {'rect': PatchesRect, 'polar': PatchesPolar, 'geo': PatchesGeo}
	if not isinstance(geometry, str): #We're gonna throw an error anyway if it's not a class or a string
		pClasses[geometry.geometry] = geometry
		geometry = geometry.geometry
	model.patches = pClasses[geometry](dim, corners=corners, offmap=offmap, **kwargs)
	def npsetter(val, item): raise RuntimeError(ï('Patch number cannot be set directly. Set the dim parameter instead.'))
	model.params['num_patch'].getter = lambda item: len(model.patches)
	model.params['num_patch'].setter = npsetter

	#Hook a positioning function or randomly position our agents
	@model.hook(prioritize=True)
	def baseAgentInit(agent, model):
		if agent.primitive == 'patch': return #Patch position is fixed
		p = model.doHooks(['baseAgentPosition', agent.primitive+'Position'], [agent, agent.model])
		if p and len(p) >= 2:
			agent.position = list(p)
			if not offmap and not agent.patch: raise ValueError(ï('Agent is not on a patch.'))
		else:
			while not offmap and not agent.patch:
				agent.position = [uniform(*model.patches.boundaries[0]), uniform(*model.patches.boundaries[1])]

	#MOVEMENT AND POSITION

	#Both agents and patches have x and y properties
	def setx(self, val): self.moveTo(val, self.position[1])
	def sety(self, val): self.moveTo(self.position[0], val)
	baseAgent.x = property(lambda self: self.position[0], setx)
	baseAgent.y = property(lambda self: self.position[1], sety)

	def move(self, x, y):
		mapx, mapy = self.model.patches.dim
		xlim, ylim = self.model.patches.boundaries
		x1 = self.position[0] + x
		y1 = self.position[1] + y

		if self.model.patches.wrap[0]:
			while x1 >= xlim[1]: x1 -= mapx
			while x1 < xlim[0]: x1 += mapx
		else:
			x1 = min(x1, xlim[1])
			x1 = max(x1, xlim[0])

		if self.model.patches.wrap[1]:
			while y1 >= ylim[1]: y1 -= mapy
			while y1 < ylim[0]: y1 += mapy
		else:
			y1 = min(y1, ylim[1])
			y1 = max(y1, ylim[0])

		if (not self.model.patches.offmap and not self.model.patches.at(x1,y1)): warnings.warn(ï('There is no patch at ({0}, {1}).').format(x1, y1), RuntimeWarning, 3)
		else: self.position = [x1, y1]

	baseAgent.move = NotPatches(move)

	#Can also take an agent or patch as a single argument
	def moveTo(self, x, y=None):
		if not isinstance(x, Number): x,y = x.position
		if x>self.model.patches.boundaries[0][1] or y>self.model.patches.boundaries[1][1] or x<self.model.patches.boundaries[0][0] or y<self.model.patches.boundaries[1][0]:
			raise IndexError(ï('Dimension is out of range.'))
		self.position = [x,y]
	baseAgent.moveTo = NotPatches(moveTo)

	#Initialize the patch container at the beginning of the model
	@model.hook(prioritize=True)
	def modelPreSetup(model):
		model.patches.clear()

	#Position our patches in the coordinate system
	@model.hook(prioritize=True)
	def patchInit(agent, model):
		model.patches.place(agent)
		agent.colorData = {}

	#Establish grid links all at once at the end.
	model.hooks.add('modelPostSetup', model.patches.neighbors, True)

	#Don't reset the visualizer if charts is already registered
	if model.visual is None or not isinstance(model.visual, Charts):
		model.useVisual(Charts)

	mapPlot = model.visual.addPlot('map', 'Map', type='agents', layout='spatial', projection=geometry if geometry=='polar' else None)
	return mapPlot

def NotPatches(function):
	"""Wraps a function to execute for all primitives except patches, which are spatially fixed."""
	def np2(self, *args, **kwargs):
		if self.primitive != 'patch': return function(self, *args, **kwargs)
		else: raise RuntimeError(ï('Patches cannot move.'))
	return np2

#===============
# PATCH GEOMETRIES
#===============

class basePatches(list, ABC):
	"""Abstract class defining the methods a coordinate system must implement. https://helipad.dev/functions/basepatches/"""

	@abstractmethod
	def revive(self, coords):
		"""Reinstates a dead patch. https://helipad.dev/functions/basepatches/revive/"""

	@abstractmethod
	def at(self, x, y):
		"""Return the patch at the specified coordinate. https://helipad.dev/functions/basepatches/at/"""

	@abstractmethod
	def neighbors(self, model):
		"""Establishes the spatial network among the patches after initialization. https://helipad.dev/functions/basepatches/neighbors/"""

	@abstractmethod
	def place(self, agent: Patch):
		"""Organize `Patch` objects within the `Patches` object. Takes a `Patch` object when created by `Agents.initialize()`, places it in the appropriate list, and assigns its position property. https://helipad.dev/functions/basepatches/place/"""

	@property
	@abstractmethod
	def boundaries(self):
		"""Maximum and minimum coordinates that agents can take, given the grid dimensions: `((xmin, xmax), (ymin, ymax))` https://helipad.dev/functions/basepatches/#boundaries"""

#Each row and column of equal length
class PatchesRect(basePatches):
	"""Defines a rectangular patch grid. https://helipad.dev/functions/patchesrect/"""
	geometry: str = 'rect'

	#Give patches and agents methods for their neighbors in the four cardinal directions
	def __init__(self, dim, wrap=True, **kwargs):
		if isinstance(wrap, bool): wrap = (wrap, wrap)
		if len(wrap) != 2: raise TypeError(ï('Invalid wrap parameter.'))
		self.wrap = wrap
		if isinstance(dim, int): dim = (dim, dim)
		if len(dim) != 2: raise TypeError(ï('Invalid dimension.'))
		self.dim = dim
		self.offmap = kwargs['offmap']
		self.corners = kwargs['corners']
		if 'noinstall' not in kwargs: RectFuncs.install()
		super().__init__([])

	def at(self, x, y): return self[round(x), round(y)]

	def neighbors(self, model):
		for patch in model.agents['patch']:
			neighbors = [(patch.right, 1), (patch.down, 1)]
			if self.corners:
				if patch.down and patch.down.right: neighbors.append((patch.down.right, self.corners))
				if patch.left and patch.left.down: neighbors.append((patch.left.down, self.corners))
			for n, weight in neighbors:
				if n: patch.edges.add(n, 'space', weight=weight)

	def place(self, agent: Patch):
		if not super().__len__(): self += [[] for i in range(self.dim[0])]
		x=0
		while len(self[x]) >= self.dim[1]: x+=1		#Find a column that's not full yet
		agent.position = (x, len(self[x]))			#Note the position
		self[x].append(agent)						#Append the agent

	@property
	def boundaries(self): return ((-0.5, self.dim[0]-0.5), (-0.5, self.dim[1]-0.5))

	def __len__(self): return self.dim[0] * self.dim[1]

	def __getitem__(self, key):
		if isinstance(key, int): return super().__getitem__(key)
		else:
			if key[1] is None: return super().__getitem__(key[0])
			if key[0] is None: return [x[key[1]] for x in self]
			patch = super().__getitem__(key[0])[key[1]]
			return patch if not patch.dead else None

	def __setitem__(self, key, val):
		if isinstance(key, int): return super().__setitem__(key, val)
		else: super().__getitem__(key[0])[key[1]] = val

	def __repr__(self): return f'<{self.__class__.__name__}: {self.dim[0]}×{self.dim[1]}>'

	def revive(self, coords):
		super().__getitem__(coords[0])[coords[1]].dead = False

#x,y placement is the same; just need to redefine patch functions and neighbors
class PatchesPolar(PatchesRect):
	"""Defines a polar-coordinate patch grid with θ×r dimensions. `x` is number around, `y` is number out. https://helipad.dev/functions/patchespolar/"""
	geometry: str = 'polar'
	wrap = (True, False) #Wrap around, but not in-to-out.

	def __init__(self, dim, **kwargs):
		PolarFuncs.install()
		super().__init__(dim, noinstall=True, **kwargs)

	def at(self, x, y): return self[floor(x), floor(y)]

	#The usual 3-4 neighbors, but if corners are on, all patches in the center ring will be neighbors
	def neighbors(self, model):
		for patch in model.agents['patch']:
			neighbors = [(patch.clockwise, 1), (patch.inward, 1)]
			if self.corners:
				neighbors += [(patch.clockwise.inward, self.corners), (patch.counterclockwise.inward, self.corners)]
				if patch.y==0:
					pc = neighbors[0][0].clockwise
					while pc.x > patch.x and pc is not patch.counterclockwise:
						neighbors.append((pc, self.corners))
						pc = pc.clockwise

			for n, weight in neighbors:
				if n: patch.edges.add(n, 'space', weight=weight)

	@property #((xmin, xmax),(ymin, ymax))
	def boundaries(self): return ((0, self.dim[0]), (0, self.dim[1]))

class PatchesGeo(basePatches):
	"""Defines a set of patches with arbitrary polygonal shapes. https://helipad.dev/functions/patchesgeo/"""
	geometry: str = 'geo'
	def __init__(self, dim=None, wrap=True, corners=True, **kwargs):
		import shapely
		self.shapely = shapely
		self.shapes = []
		self.corners = corners
		if isinstance(wrap, bool): wrap = (wrap, wrap)
		if len(wrap) != 2: raise TypeError(ï('Invalid wrap parameter.'))
		self.wrap = wrap
		self.offmap = kwargs['offmap']

		def vertices(patch): return patch.polygon.exterior.xy
		def area(patch): return patch.polygon.area
		def center(patch): return (patch.polygon.centroid.x, patch.polygon.centroid.y) if hasattr(patch, 'polygon') else None
		def agentsOn(patch):
			agents = []
			for a in patch.model.agents.all:
				if a.primitive == 'patch': continue
				if patch.polygon.covers(self.shapely.Point(*a.position)): agents.append(a)
			return agents
		for prop in [vertices, area, center, agentsOn]: setattr(Patch, prop.__name__, property(prop))
		Patch.position = property(center)
		RectFuncs.install(patches=False)

	def __getitem__(self, val):
		if isinstance(val, str): patch = [p for p in self if p.name==val][0]
		else: patch = super().__getitem__(val)
		return patch if not patch.dead else None

	def __len__(self): return len(self.shapes)

	@property
	def names(self):
		return [p.name for p in self if p.name is not None]

	@property
	def boundaries(self):
		bounds = [s.shape.bounds for s in self.shapes]
		minx, miny, maxx, maxy = bounds[0] if bounds else (0,0,0,0)
		for b in bounds:
			minx, maxx = min(minx, b[0]), max(maxx, b[2])
			miny, maxy = min(miny, b[1]), max(maxy, b[3])
		return ((minx, maxx), (miny, maxy))

	@property
	def dim(self):
		bounds = self.boundaries
		return (bounds[0][1]-bounds[0][0], bounds[1][1]-bounds[1][0])

	def add(self, shape, name=None):
		"""Add a polygonal patch to the map. https://helipad.dev/functions/patchesgeo/add/"""
		if isinstance(shape, list): shape = self.shapely.Polygon(shape)
		elif isinstance(shape, self.shapely.MultiPolygon):
			warnings.warn(ï('MultiPolygons are not supported as patches. Taking the first polygon…'), RuntimeWarning, 2)
			shape = shape.geoms[0]
		if name and name in self.names: raise KeyError(ï('Patch with name \'{0}\' already exists.').format(name))
		item = Item(shape=shape, name=name, borders=[], corners=[]) #Have to store this as a wrapper item because Shapely objects are immutable

		#Ensure no overlap and calculate neighbors
		for i,p in enumerate(self.shapes):
			intersection = shape.intersection(p.shape)
			if not intersection.is_empty:
				if isinstance(intersection, (self.shapely.LineString, self.shapely.MultiLineString)): item.borders.append(p)
				elif isinstance(intersection, self.shapely.Point): item.corners.append(p)
				else:
					pn = f'\'{p.name}\'' if p.name else i
					raise ValueError(ï('Polygon {0} overlaps existing patch {1}.').format(name, pn))

		self.shapes.append(item)

	def at(self, x, y):
		pt = self.shapely.Point(x,y)
		for p in self:
			if p.polygon.covers(pt): return p if not p.dead else None

	def neighbors(self, model):
		for p in self.shapes:
			for e in p.borders: p.patch.edges.add(e.patch, 'space')
			if self.corners:
				for e in p.corners: p.patch.edges.add(e.patch, 'space', weight=self.corners)

	def place(self, agent: Patch):
		shape = self.shapes[super().__len__()]
		agent.polygon, agent.name = shape.shape, shape.name
		shape.patch = agent
		self.append(agent)

	def revive(self, index):
		if isinstance(index, str): [p for p in self if p.name==index][0].dead = False
		else: super().__getitem__(index).dead = False

#===============
# COORDINATE SYSTEMS
#===============

class RectFuncs:
	"""Installs distance, neighbor, and angle functions for rectangular coordinates into agent objects. This is a static class and should not be instantiated."""
	@staticmethod
	def install(patches=True, agents=True, allagents=True):
		if patches:
			for p in ['up', 'right', 'down', 'left', 'agentsOn', 'center', 'area', 'vertices']:
				setattr(Patch, p, property(getattr(RectFuncs,p)))
		if agents:
			for f in ['moveUp', 'moveRight', 'moveDown', 'moveLeft', 'forward', 'orientTo']:
				setattr(baseAgent, f, NotPatches(getattr(RectFuncs,f)))
		if allagents:
			for h in ['_offset', 'distanceFrom']:
				setattr(baseAgent, h, getattr(RectFuncs,h))

	def up(patch):
		if patch.y==0 and not patch.model.patches.wrap[1]: return None
		return patch.model.patches[patch.x, patch.y-1 if patch.y > 0 else patch.model.patches.dim[1]-1]

	def right(patch):
		if patch.x>=patch.model.patches.dim[0]-1 and not patch.model.patches.wrap[0]: return None
		return patch.model.patches[patch.x+1 if patch.x < patch.model.patches.dim[0]-1 else 0, patch.y]

	def down(patch):
		if patch.y>=patch.model.patches.dim[1]-1 and not patch.model.patches.wrap[1]: return None
		return patch.model.patches[patch.x, patch.y+1 if patch.y < patch.model.patches.dim[1]-1 else 0]

	def left(patch):
		if patch.x==0 and not patch.model.patches.wrap[0]: return None
		return patch.model.patches[patch.x-1 if patch.x > 0 else patch.model.patches.dim[0]-1, patch.y]

	def agentsOn(patch):
		for prim, lst in patch.model.agents.items():
			if prim=='patch': continue
			yield from [a for a in lst if patch.x-0.5<=a.x<patch.x+0.5 and patch.y-0.5<=a.y<patch.y+0.5]

	def center(patch): return patch.position
	def area(patch): return 1
	def vertices(patch):
		pp = patch.position
		return ((pp[0]-0.5, pp[1]-0.5), (pp[0]-0.5, pp[1]+0.5), (pp[0]+0.5, pp[1]+0.5), (pp[0]+0.5, pp[1]-0.5))

	def moveUp(agent): agent.move(0, -1)
	def moveRight(agent): agent.move(1, 0)
	def moveDown(agent): agent.move(0, 1)
	def moveLeft(agent): agent.move(-1, 0)
	def forward(self, steps=1):
		"""Move forward `steps` steps in the direction of `agent.orientation`."""
		self.move(steps * cos(self.rads-pi/2), steps * sin(self.rads-pi/2))

	def orientTo(agent, x, y=None):
		if not isinstance(x, Number): x,y = x.position
		difx, dify = agent._offset(x, y)
		agent.rads = atan2(dify,difx) + 0.5*pi #atan calculates 0º pointing East. We want 0º pointing North

	#Internal function  to calculate x,y distances s.t. wrap constraints for distance and orientation purposes
	def _offset(agent, x, y):
		difx, dify = x - agent.x, y - agent.y
		if agent.model.patches.wrap[0]:
			crossx = difx - copysign(agent.model.patches.dim[0], difx)
			if abs(crossx) < abs(difx): difx = crossx
		if agent.model.patches.wrap[1]:
			crossy = dify - copysign(agent.model.patches.dim[1], dify)
			if abs(crossy) < abs(dify): dify = crossy
		return (difx, dify)

	def distanceFrom(agent, agent2):
		difx, dify = agent._offset(agent2.x, agent2.y)
		return sqrt(difx**2 + dify**2)

class PolarFuncs:
	"""Installs distance, neighbor, and angle functions for polar coordinates into agent objects. This is a static class and should not be instantiated."""
	@staticmethod
	def install(patches=True, agents=True, allagents=True):
		if patches:
			for p in ['inward', 'outward', 'clockwise', 'counterclockwise', 'agentsOn', 'center', 'area', 'vertices']:
				setattr(Patch, p, property(getattr(PolarFuncs,p)))
		if agents:
			for f in ['moveInward', 'moveOutward', 'moveClockwise', 'moveCounterclockwise', 'forward', 'orientTo']:
				setattr(baseAgent, f, NotPatches(getattr(PolarFuncs,f)))
		if allagents:
			baseAgent.distanceFrom = PolarFuncs.distanceFrom

	def inward(patch):
		if patch.y==0: return None
		return patch.model.patches[patch.x, patch.y-1]

	def outward(patch):
		if patch.y>=patch.model.patches.dim[1]-1: return None
		return patch.model.patches[patch.x, patch.y+1]

	def clockwise(patch):
		return patch.model.patches[patch.x+1 if patch.x < patch.model.patches.dim[0]-1 else 0, patch.y]

	def counterclockwise(patch):
		return patch.model.patches[patch.x-1 if patch.x > 0 else patch.model.patches.dim[0]-1, patch.y]

	def agentsOn(patch):
		for prim, lst in patch.model.agents.items():
			if prim=='patch': continue
			yield from [a for a in lst if floor(a.x)==patch.x and floor(a.y)==patch.y]

	def center(patch): return (patch.position[0]+0.5, patch.position[1]+0.5)
	def area(patch): return (1/patch.model.patches.dim[0])*pi*((patch.position[1]+1)**2-patch.position[1]**2)
	def vertices(patch):
		pp = patch.position
		return (pp, (pp[0], pp[1]+1), (pp[0]+1, pp[1]+1), (pp[0]+1, pp[1]))

	def moveInward(agent): agent.move(0, -1)
	def moveOutward(agent): agent.move(0, 1)
	def moveClockwise(agent): agent.move(1, 0)
	def moveCounterclockwise(agent): agent.move(-1, 0)
	def forward(agent, steps=1):
		"""Move forward `steps` steps in the direction of `agent.orientation`."""
		#2π-self.rads to make it go clockwise
		th1 = 2*pi-(2*pi/agent.model.patches.dim[0] * agent.x)
		newth = th1 + atan2(steps*sin(2*pi-agent.rads-th1), agent.y+steps*cos(2*pi-agent.rads-th1))
		newx = (2*pi-newth)/(2*pi/agent.model.patches.dim[0])
		if newx > agent.model.patches.dim[0]: newx -= agent.model.patches.dim[0]
		newr = sqrt(agent.y**2 + steps**2 + 2*agent.y*steps*cos(2*pi-agent.rads-th1))
		agent.move(newx-agent.x, newr-agent.y)

	def orientTo(agent, x, y=None):
		if not isinstance(x, Number): x,y = x.position
		th1 = 2*pi-(2*pi/agent.model.patches.dim[0] * agent.x)
		th2 = 2*pi-(2*pi/agent.model.patches.dim[0] * x)
		agent.rads = pi - atan2(agent.y * sin(th1) - y * sin(th2), agent.y * cos(th1) - y * cos(th2))

	def distanceFrom(agent, agent2):
		th1 = agent.x * 2*pi /agent.model.patches.dim[0]
		th2 = agent2.x * 2*pi /agent.model.patches.dim[0]
		return sqrt(agent.y**2 + agent2.y**2 - 2*agent.y*agent2.y*cos(th1-th2))


# A model of the long-run cyclical dynamics of urbanization and human capital.

from math import sqrt, floor
from numpy import diff, random, log
from helipad import Helipad
heli = Helipad()

#================
# CONFIGURATION
#================

heli.agents.addBreed('urban', '#CC0000')
heli.agents.addBreed('rural', '#00CC00')

#Constrain the parameter space to values resulting in H* = 100
def constrain(model, var, val):
	if var=='city':
		if val: model.params.hide('rent')
		else: model.params.show('rent')
		#model.params['rent'].disabled(val)
	elif model.param('city'):
		if var=='fixed':
			model.params['rent'].enable() #Tkinter won't let you update the value of a disabled widget…
			model.param('rent', .04+.037*val)
			model.params['rent'].disable()
		# if var=='rent':
		# 	model.params['fixed'].enable() #Tkinter won't let you update the value of a disabled widget…
		# 	model.param('fixed', 24.7777*val-.9284659)
		# 	model.params['fixed'].disable()

heli.params.add('city', 'City?', 'check', True, desc='Whether agents have the possibility of moving to the city', runtime=False, callback=constrain)
heli.params.add('lockH', 'Lock human capital', 'check', False, desc='Maintains the distribution of human capital when checked')
heli.params.add('breedThresh', 'Breeding Threshold (φ)', 'slider', dflt=20, opts={'low':5, 'high': 500, 'step': 5}, desc='Proportional to the minimum wealth necessary to breed')
heli.params.add('movecost', 'Moving Cost (ω)', 'slider', dflt=10, opts={'low':0, 'high': 150, 'step': 1}, desc='Cost incurred by moving location')
heli.params.add('deathrate', 'Death Rate (θ)', 'slider', dflt=0.005, opts={'low':0, 'high': 0.05, 'step': 0.001})
heli.params.add('rent', 'Variable cost (ρ)', 'slider', dflt=.04, opts={'low':0.01, 'high': 0.1, 'step': 0.01}, desc='Per-period cost-of-living, proportional to human capital', callback=constrain)
heli.params.add('fixed', 'Fixed cost (χ)', 'slider', dflt=.4, opts={'low':0, 'high': 1, 'step': 0.1}, desc='Per-period cost-of-living', callback=constrain)

heli.name = 'Cities'
heli.stages = 2
heli.agents.order = 'linear'

popfactor = 10 #Multiplies the productivity and therefore the equilibrium population, without changing anything else
burnout = 2000 #The end period, once it's equilibrated, over which to average the values

class Land():
	def __init__(self, loc):
		self.loc = loc
		self.input = 0
		self.lastInput = 0
		self.product = 0

	def produce(self):
		self.product = popfactor*log(self.input) #if self.loc=='rural' else sqrt(self.input)	# = ln∑P (eq. 2)
		self.lastInput = self.input
		self.input = 0 #Reset productivity at the end of each period
		return self.product

	def pop(self, model): return len(model.agents['agent'][self.loc])

	#The add argument calculates the productivity as if the agent were already there, if he's not
	def agentProd(self, H, add=False):
		if self.loc=='rural': return sqrt(150) * H
		else:
			if add:
				return 1/sqrt(self.pop(heli)+1) * (heli.data.getLast('hsum')+H)
				# return (heli.data.getLast('urbanH')*p+H)/(p+1) * log(heli.data.getLast('hsum')+H) #wtf was I doing here?
			else: return 1/sqrt(self.pop(heli)) * heli.data.getLast('hsum')
			# else: return heli.data.getLast('urbanH')*log(heli.data.getLast('hsum'))

	def expWage(self, H):
		prod = self.agentProd(H, True)
		potentialprod = log(self.lastInput+prod) #if self.loc=='rural' else sqrt(self.lastInput+prod)
		return potentialprod *  prod/(self.lastInput+prod) #Your share of the product

heli.land = {k: Land(k) for k in ['urban', 'rural']}
heli.params['num_agent'].type = 'hidden'

#================
# AGENT BEHAVIOR
#================

@heli.hook
def CpanelPostInit(cpanel):
	constrain(cpanel.model, 'city', True)

#This is here to make sure that the agents param gets reset at the beginning of each run
#Otherwise the parameter persists between runs
@heli.hook
def modelPreSetup(model):
	model.movers = {b:0 for b in heli.agents['agent'].breeds}
	model.births = {b:0 for b in heli.agents['agent'].breeds}
	model.deaths = 0
	for b in heli.agents['agent'].breeds:
		setattr(model, 'moverate'+b, 0)
		setattr(model, 'birthrate'+b, 0)
	model.deathrate = 0

	#Mark the different phases of the Malthusian model
	model.events.clear()
	if not model.param('city'):
		model.param('num_agent', 150)
		@heli.event
		def exp1(model):	return sum(diff(model.data.getLast('ruralPop', 20))) > 3
		@heli.event
		def stab2(model):	return model.events['exp1'].triggered and model.t > model.events['exp1'].triggered+1000 and sum(diff(model.data.getLast('ruralH', 3000))) < 0
		@heli.event
		def end3(model): return model.events['stab2'].triggered and model.t >= model.events['stab2'].triggered + burnout

	#Start with the equilibrium population
	else: model.param('num_agent', floor(2.55-1.69*model.param('fixed'))*popfactor)

@heli.hook
def agentInit(agent, model):
	agent.H = random.normal(100, 15) if model.param('city') else 10 #Human capital
	agent.prod = 0
	agent.wealth = model.param('breedThresh') + model.param('rent')*10
	agent.lastWage = 0
	agent.expwage = 0

#Distribute product in proportion to input
#modelStep executes each stage **before** any agent step functions
@heli.hook
def modelStep(model, stage):
	if stage==2:
		for l in model.land.values():
			inp = l.input								# = ∑P
			Y = l.produce()								# = ln∑P (eq. 2)
			for a in model.agents['agent'][l.loc]:
				a.lastWage = a.prod/inp * Y	# = P/∑P * Y (eq. 3)
				a.wealth += a.lastWage

@heli.hook
def agentStep(agent, model, stage):
	if stage==1:
		if model.param('city'):
			otherloc = 'urban' if agent.breed == 'rural' else 'rural'
			agent.expwage = model.land[otherloc].expWage(agent.H)

			#Decide whether or not to move
			mvc = model.param('movecost')
			if agent.expwage > agent.lastWage*1.2 and agent.wealth > mvc and model.t > 2:
			# if agent.prod[otherloc]-mvc > agent.prod[agent.breed] and agent.wealth > mvc: #and model.t > 10000:
				# print('T=',model.t,', HC',agent.H,':',agent.breed,'wage=',agent.lastWage,',',otherloc,'wage=',otherwage)
				agent.wealth -= mvc
				model.movers[agent.breed] += 1
				agent.breed = otherloc

		agent.prod = model.land[agent.breed].agentProd(agent.H)
		model.land[agent.breed].input += agent.prod #Work

		#Reproduce
		randn = random.normal(1, 0.2)
		if random.random() < model.param('deathrate'): agent.die()
		elif agent.wealth > model.param('breedThresh') * (randn if agent.breed=='rural' else agent.H/4):
			child = agent.reproduce(
				inherit=['H', ('wealth', lambda w: w[0]/2)],
				mutate={'H': 14.25} if not model.param('lockH') else {} #See footnote 3 for derivation
			)
			agent.wealth -= agent.wealth/2 + 1 #-= breedThresh #Fixed cost
			model.births[agent.breed] += 1
			if child.H < 2: child.die()

	#Get paid in modelStep, then pay rent
	elif stage==2:
		agent.wealth -= model.param('rent') * agent.H + model.param('fixed')
		if agent.wealth <= 0: agent.die()
		return

#Organize some data and pause if all agents are dead
@heli.hook
def modelPostStep(model):
	if len(model.agents['agent']) == 0: model.stop()
	else:
		for b in model.agents['agent'].breeds:
			pop = len(model.agents['agent'][b])
			if pop > 0:
				setattr(model,'moverate'+b, model.movers[b]/pop)
				setattr(model,'birthrate'+b, model.births[b]/pop)
				model.movers[b] = 0
				model.births[b] = 0
		model.deathrate = model.deaths/len(model.agents['agent'])
		model.deaths = 0

@heli.hook
def decideBreed(aId, choices, model):
	return 'rural'

@heli.hook
def agentDie(agent): heli.deaths += 1

#================
# REPORTERS AND PLOTS
#================

from helipad.visualize import TimeSeries
viz = heli.useVisual(TimeSeries)

# returns ln(∑H²)
@heli.reporter
def hsum(model, loc='urban'):
	upop = model.agents['agent'][loc]
	return sum(a.H for a in upop) if len(upop) > 0 else 1

def perCapGdp(model, loc):
	def tmp(model):
		pop = model.land[loc].pop(model)
		if pop==0: return 0
		else: return model.land[loc].product/pop
	return tmp

viz.addPlot('pop', 'Population', 1, logscale=True)
viz.addPlot('hcap', 'Human Capital', 2)
viz.addPlot('wage', 'Wage', 3)
viz.addPlot('wealth', 'Wealth', 4)
viz.addPlot('rates', 'Rates', 5, logscale=True)
heli.data.addReporter('theta', heli.data.modelReporter('deathrate'), smooth=0.99)
viz['rates'].addSeries('theta', 'Death Rate', '#CCCCCC')

for breed, d in heli.agents['agent'].breeds.items():
	heli.data.addReporter(breed+'Pop', heli.land[breed].pop)
	heli.data.addReporter(breed+'H', heli.data.agentReporter('H', 'agent', breed=breed, percentiles=[25,75]))
	heli.data.addReporter(breed+'Wage', heli.data.agentReporter('lastWage', 'agent', breed=breed, percentiles=[25,75]))
	heli.data.addReporter(breed+'ExpWage', heli.data.agentReporter('expwage', 'agent', breed=breed, percentiles=[25,75]))
	heli.data.addReporter(breed+'Wealth', heli.data.agentReporter('wealth', 'agent', breed=breed, percentiles=[25,75]))
	heli.data.addReporter(breed+'moveRate', heli.data.modelReporter('moverate'+breed), smooth=0.99)
	heli.data.addReporter(breed+'birthrate', heli.data.modelReporter('birthrate'+breed), smooth=0.99)
	viz['pop'].addSeries(breed+'Pop', breed.title()+' Population', d.color)
	viz['hcap'].addSeries(breed+'H', breed.title()+' Human Capital', d.color)
	viz['wage'].addSeries(breed+'Wage', breed.title()+' Wage', d.color)
	viz['wage'].addSeries(breed+'ExpWage', breed.title()+' Expected Wage', d.color2, visible=False)
	viz['wealth'].addSeries(breed+'Wealth', breed.title()+' Wealth', d.color)
	viz['rates'].addSeries(breed+'moveRate', breed.title()+' Moveaway Rate', d.color2)
	viz['rates'].addSeries(breed+'birthrate', breed.title()+' Birthrate', d.color)

heli.launchCpanel()

#================
# PARAM SWEEP & ANALYSIS
#================

# for p in viz.values(): p.active(False) #Disable plots so Helipad doesn't try to update the visuals during param sweep
# #Remove superfluous columns
# @heli.hook
# def saveCSV(data, model):
# 	for c in data:
# 		if 'urban' in c or c in ['city', 'hsum'] or 'utility' in c or 'moveRate' in c or 'Unnamed' in c or 'pctile' in c:
# 			del data[c]
# 	print('Finished at', len(data))
# 	return data
#
# #Set up parameter sweep
# folder = 'CSVs'
# heli.param('csv', folder+'/malthusian')
# @heli.hook('modelPreSetup', prioritize=True)
# def turnOffCity(model): model.param('city', False)
# heli.param('stopafter', 'end3')
# results = heli.paramSweep(['rent', 'fixed'])
#
# #Post-sweep analysis
# import os, pandas
# times = [{e.name: e.triggered for e in run.events} for run in results]
# files = [f for f in os.listdir(folder) if '.csv' in f]
# analysis = []
# i=0
# for f in files:
# 	data = pandas.read_csv(folder+'/'+f)
#
# 	#Summary statistics
# 	analysis.append({
# 		'rent': data['rent'][0],
# 		'fixed': data['fixed'][0],
# 		'h': mean(data['ruralH'][-burnout:]),
# 		'pop': mean(data['ruralPop'][-burnout:])/popfactor
# 	}|times[i])
# 	i+=1
#
# analysis = pandas.DataFrame(analysis)
# analysis.to_csv(folder+'/analysis.csv')


"""
Helipad is an agent-based modeling framework with a hook-based architecture and powerful visualization capabilities. Import and instantiate the `Helipad` class to set up and launch a model. See https://helipad.dev for API documentation.
"""

from helipad.model import Helipad, MultiLevel
from helipad.agent import Agent, baseAgent
from helipad import utility

__version__ = "1.6.2"


"""
The `Agent` class and the `Agents` container. This module can be imported to extend `baseAgent` to create new primitives.
"""

import warnings
from random import choice, randint, shuffle
from math import degrees, radians, pi
import numpy as np
import pandas
from helipad.helpers import ï, funcStore, Color, Item, isNotebook

#Basic agent functions. This class should not be instantiated directly; instead it should be
#subclassed by a class corresponding to a primitive and registered with Helipad.addPrimitive().
#See below, the Agent() class for a minimal example.
class baseAgent:
	"""A basic agent class that can be subclassed to create new primitives (see https://helipad.dev/functions/agents/addprimitive/). https://helipad.dev/functions/baseagent/"""
	angleUnit: str = 'deg'
	fixed: bool = False
	overdraft: str = 'continue-silent'

	#==================
	# BASIC METHODS
	#==================

	def __init__(self, breed: str, aId: int, model):
		self.breed = breed
		self.id = int(aId)
		self.model = model
		self.age = 0
		self.dead = False
		self.stocks = Stocks(breed, model.goods)
		self.edges = Edges(self)
		self.utils = 0
		if not hasattr(self, 'position'): self.position = None #Overridden in spatial init
		self.currentDemand = {g:0 for g in model.goods.keys()}
		self.rads = 0

		#For multi-level models
		#Has to be a static property since we're checking as the object initializes
		if hasattr(super(), 'runInit'): super().__init__()

		self.model.doHooks(['baseAgentInit', self.primitive+'Init'], [self, self.model])

	def __repr__(self): return f'<{self.__class__.__name__} {self.id}>'

	def step(self, stage: int):
		"""Runs for each agent in each period and increments the agent's age by 1. Should not be called directly, but should be hooked using `agentStep` or `baseAgentStep`. https://helipad.dev/functions/baseagent/step/"""
		self.model.doHooks(['baseAgentStep', self.primitive+'Step'], [self, self.model, stage])
		if hasattr(super(), 'runInit'): super().step(stage) #For multi-level models
		if stage == self.model.stages: self.age += 1

	#==================
	# ECONOMIC METHODS
	#==================

	def trade(self, partner, good1: str, amt1, good2: str, amt2):
		"""Exchange `amt1` of `good1` with `partner` for `amt2` of `good2`, and record the demand for each. Requires at least two goods to have been registered, but does not require a monetary good. Negative amounts are allowed, which reverses the direction. https://helipad.dev/functions/baseagent/trade/"""
		go = True
		self.model.doHooks('preTrade', [self, partner, good1, amt1, good2, amt2])

		#Budget constraints. Hold price constant if hit
		message = None
		prim1, prim2 = self.primitive.title(), partner.primitive.title()
		if amt2 != 0: price = amt1 / amt2
		if self.overdraft != 'allow':
			if amt1 > self.stocks[good1]:
				message = ï('{0} {1} does not have sufficient {2} to give {3} {4}.').format(prim1, self.id, good1, prim2, partner.id)
				if 'continue' in self.overdraft:
					message += ï(' Continuing with available {0} of {1}…').format(good1, self.stocks[good1])
					amt1 = self.stocks[good1]
					if amt2 != 0: amt2 = amt1 / price
			elif -amt1 > partner.stocks[good1]:
				message = ï('{0} {1} does not have sufficient {2} to give {3} {4}.').format(prim2, partner.id, good1, prim1, self.id)
				if 'continue' in self.overdraft:
					message += ï(' Continuing with available {0} of {1}…').format(good1, partner.stocks[good1])
					amt1 = -partner.stocks[good1]
					if amt2 != 0: amt2 = amt1 / price
			if amt2 > partner.stocks[good2]:
				message = ï('{0} {1} does not have sufficient {2} to give {3} {4}.').format(prim2, partner.id, good2, prim1, self.id)
				if 'continue' in self.overdraft:
					message += ï(' Continuing with available {0} of {1}…').format(good2, partner.stocks[good2])
					amt2 = partner.stocks[good2]
					amt1 = price * amt2
			elif -amt2 > self.stocks[good2]:
				message = ï('{0} {1} does not have sufficient {2} to give {3} {4}.').format(prim1, self.id, good2, prim2, partner.id)
				if 'continue' in self.overdraft:
					message += ï(' Continuing with available {0} of {1}…').format(good2, self.stocks[good2])
					amt2 = -self.stocks[good2]
					amt1 = price * amt2

			if message is not None:
				if self.overdraft == 'stop': raise ValueError(message)
				if 'fail' in self.overdraft:
					go = False
					message += ï(' Cancelling trade…')
				elif 'warn' in self.overdraft: warnings.warn(message, None, 2)

		if go:
			self.stocks[good1] -= amt1
			partner.stocks[good1] += amt1
			self.stocks[good2] += amt2
			partner.stocks[good2] -= amt2

			#Record demand
			if amt1 > 0: partner.currentDemand[good1] += amt1
			else: self.currentDemand[good1] -= amt1
			if amt2 > 0: self.currentDemand[good2] += amt2
			else: partner.currentDemand[good2] -= amt2

		self.model.doHooks('postTrade', [self, partner, good1, amt1, good2, amt2])

	def buy(self, partner, good: str, q, p):
		"""Purchase `q` of `good` from `partner` at a per-unit price of `p` in terms of the monetary good. Returns the quantity actually sold, which is the same as `q` unless there's a shortage. https://helipad.dev/functions/baseagent/buy/"""
		if self.model.goods.money is None: raise RuntimeError(ï('{} requires a monetary good to be specified.').format('Agent.buy()'))
		qp = self.model.doHooks('buy', [self, partner, good, q, p])
		if qp is not None: q, p = qp

		before = self.stocks[good]
		self.trade(partner, self.model.goods.money, p*q, good, q)
		return self.stocks[good] - before

	def pay(self, recipient, amount):
		"""Unilaterally transfer `amount` of the monetary good to `recipient`, or from `recipient` to the agent if `amount<0`. Returns the amount actually paid, which will be equal to `amount` unless the agent's budget constraint is hit. https://helipad.dev/functions/baseagent/pay/"""
		if self.model.goods.money is None: raise RuntimeError(ï('{} requires a monetary good to be specified.').format('Agent.pay()'))
		go = True

		#Do hooks before budget constraints
		amount_ = self.model.doHooks('pay', [self, recipient, amount, self.model])
		if amount_ is not None: amount = amount_

		#Budget constraints
		prim1, prim2 = self.primitive.title(), recipient.primitive.title()
		message = None
		if self.overdraft != 'allow':
			if amount > self.balance:
				message = ï('{0} {1} does not have sufficient funds to pay {2} {3}.').format(prim1, self.id, prim2, recipient.id)
				if self.overdraft == 'stop': raise ValueError(message)
				if 'continue' in self.overdraft:
					amount = self.balance
					message += ï(' Continuing with available balance of {}…').format(self.balance)
			elif -amount > recipient.balance:
				message = ï('{0} {1} does not have sufficient funds to pay {2} {3}.').format(prim2, recipient.id, prim1, self.id)
				if 'continue' in self.overdraft:
					amount = -recipient.balance
					message += ï(' Continuing with available balance of {}…').format(recipient.balance)

			if message is not None:
				if self.overdraft == 'stop': raise ValueError(message)
				if 'fail' in self.overdraft:
					go = False
					message += ï(' Cancelling trade…')
				if 'warn' in self.overdraft:
					warnings.warn(message, None, 2)

		if go and amount:
			recipient.stocks[self.model.goods.money] += amount
			self.stocks[self.model.goods.money] -= amount
			return amount
		else: return 0

	@property
	def balance(self):
		"""If a monetary good is registered, the agent's holdings of the monetary good. Equivalent to `agent.stocks[model.goods.money]`. https://helipad.dev/functions/baseagent/#balance"""
		if self.model.goods.money is None: raise RuntimeError(ï('Balance checking requires a monetary good to be specified.'))
		bal = self.stocks[self.model.goods.money]
		bal_ = self.model.doHooks('checkBalance', [self, bal, self.model])
		if bal_ is not None: bal = bal_

		return bal

	#==================
	# GENETIC METHODS
	#==================

	def reproduce(self, inherit: list=[], mutate: dict={}, partners: list=[]):
		"""Spawn a new agent, inheriting specified properties from the parent agent(s).
		
		Values in `inherit` can either be a property name (in which case the new agent inherits the mean of the parents' values), or a tuple, the first element of which is a property name, and the second is a string representing how to merge them. Possible values are `'mean'` (default for numeric values), `'first'` (default for non-numeric values), `'last'`, `'gmean'`, `'random'`, and `'sum'`. The second value can also take a function, which receives a list of values from the parents and returns a value for the child.
		
		Keys of `mutate` correspond to property names, which will mutate either a property value retrieved from `inherit` or the initial value otherwise. Dict values can be either (1) a function (which takes a value and returns a value), (2) a std dev by which to mutate the value in a normal distribution with mean of the initial value, or (3) a tuple, the first item of which is a stdev and the second is either `'log'` or `'linear'` (i.e. to sample from a lognormal distribution).

		https://helipad.dev/functions/baseagent/reproduce/"""
		if self.fixed: raise NotImplementedError(ï('Fixed primitives cannot reproduce.'))

		maxid = 0
		for a in self.model.agents.all:
			maxid = max(maxid, a.id)
		newagent = type(self)(self.breed, maxid+1, self.model)

		parents = [self] + partners
		for a in inherit:
			stat = None
			if isinstance(a, tuple): a, stat = a
			v = [getattr(p,a) for p in parents if hasattr(p,a)] #List of values, filtering those without
			if not v: continue

			#Default statistic if unspecified. 'mean' for numbers, and 'first' for non-numbers.
			if stat is None:
				stat = 'mean' if isinstance(v[0], (int, float, complex)) and not isinstance(v[0], bool) else 'first'

			if stat=='mean': n = np.mean(v)
			elif stat=='sum': n = sum(v)
			elif stat=='gmean': n = np.exp(np.log(v).sum()/len(v))
			elif stat=='first': n = v[0]
			elif stat=='last': n = v[-1]
			elif stat in ('rand', 'random'): n = choice(v)
			elif stat=='max': n = max(v)
			elif stat=='min': n = min(v)
			elif callable(stat): n = stat(v)
			else: raise ValueError(ï('Invalid statistic {}.').format(stat))

			setattr(newagent, a, n)

		#Mutate variables
		for k,v in mutate.items():
			if callable(v): newval = v(getattr(newagent, k))
			else:
				if isinstance(v, tuple): v, scale = v
				else: scale = 'linear'

				if scale=='log': newval = np.random.lognormal(np.log(getattr(newagent, k)), v)
				else: newval = np.random.normal(getattr(newagent, k), v)
			setattr(newagent, k, newval)

		newagent.id = maxid+1
		for p in parents:
			p.edges.add(newagent,'lineage', True) #Keep track of parent-child relationships
		self.model.agents[self.primitive].append(newagent)

		self.model.doHooks(['baseAgentReproduce', self.primitive+'Reproduce'], [parents, newagent, self.model])
		return newagent

	def die(self, updateGUI: bool=True):
		"""Remove the agent from the model's list of active agents and cut the agent's edges. https://helipad.dev/functions/baseagent/die/"""
		if self.fixed: raise NotImplementedError(ï('Fixed primitives cannot die.'))
		self.model.agents[self.primitive].remove(self)
		for edge in self.edges.all: edge.cut()
		self.dead = True
		self.model.doHooks(['baseAgentDie', self.primitive+'Die'], [self])

	@property
	def parent(self):
		"""The agent (if haploid) or a list of agents (if polyploid) from which the current agent was spawned. See `agent.reproduce()`. https://helipad.dev/functions/baseagent/#parent"""
		p = self.inbound('lineage', obj='agent')
		if len(p)==0: return None
		elif len(p)==1: return p[0]
		else: return p

	@property
	def children(self):
		"""A list of agents spawned from the current agent. See `agent.reproduce()`. https://helipad.dev/functions/baseagent/#children"""
		return [edge.partner(self) for edge in self.outbound('lineage')]

	#==================
	# NETWORK METHODS
	# All deprecated in Helipad 1.6, remove in Helipad 1.8
	#==================

	def newEdge(self, partner, kind='edge', direction=None, weight=1):
		"""Create a network connection with `partner`. This method is deprecated; use `Agent.edges.add()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Agent.newEdge()', 'Agent.edges.add()'), FutureWarning, 2)
		return self.edges.add(partner, kind, direction, weight)

	def outbound(self, kind='edge', undirected=False, obj='edge'):
		"""Return a list of outbound edges. This method is deprecated; use `Agent.edges.outbound()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Agent.outbound()', 'Agent.edges.outbound()'), FutureWarning, 2)
		return self.edges.outbound(kind, undirected, obj)

	def inbound(self, kind='edge', undirected=False, obj='edge'):
		"""Return a list of inbound edges. This method is deprecated; use `Agent.edges.inbound()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Agent.inbound()', 'Agent.edges.inbound()'), FutureWarning, 2)
		return self.edges.inbound(kind, undirected, obj)

	def edgesWith(self, partner, kind='edge'):
		"""Return a list of connections with `partner`. This method is deprecated; use `Agent.edges.With()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Agent.edgesWith()', 'Agent.edges.With()'), FutureWarning, 2)
		return self.edges.With(partner, kind)

	@property
	def alledges(self):
		"""A list of all the agent's network connections, organized by network kind. This method is deprecated; use `Agent.edges` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Agent.allEdges', 'Agent.edges.all'), FutureWarning, 2)
		return self.edges.all

	#==================
	# OTHER METHODS
	#==================

	#Agent.orientation reports and sets degrees or radians, depending on Agent.angleUnit.
	#Agent.rads always reports radians and is not documented.
	@property
	def orientation(self):
		"""The orientation of the agent on the spatial grid. This property returns and can be set in either degrees or radians, depending on the value of `baseAgent.angleUnit`. https://helipad.dev/functions/baseagent/#orientation"""
		if self.primitive == 'patch': return None
		return degrees(self.rads) if 'deg' in self.angleUnit else self.rads

	@orientation.setter
	def orientation(self, val):
		if self.primitive == 'patch': raise RuntimeError(ï('Patches cannot rotate.'))
		if 'deg' in self.angleUnit: val = radians(val)
		while val >= 2*pi: val -= 2*pi
		while val < 0: val += 2*pi
		self.rads = val

	def rotate(self, angle):
		"""Rotate the agent's orientation clockwise by `angle`, which will be understood as either degrees or radians depending on the value of `baseAgent.angleUnit`. https://helipad.dev/functions/baseagent/rotate/"""
		self.orientation += angle

	@property
	def patch(self):
		"""The patch under the agent's current position. https://helipad.dev/functions/baseagent/#patch"""
		if self.position is not None: return self.model.patches.at(*self.position)

	def transfer(self, dest):
		"""In a multi-level model, move the agent to a different instance of `MultiLevel` or the top-level model."""
		origin = self.model
		dest.agents[self.primitive].append(self)
		self.model = dest
		origin.agents[self.primitive].remove(self)
		self.model.doHooks(['baseAgentMove', self.primitive+'Move'], [self, origin, dest])

#The default agent class corresponding to the 'agent' primitive.
class Agent(baseAgent):
	"""The default agent class. https://helipad.dev/functions/agent/"""

#For spatial models
class Patch(baseAgent):
	"""A spatially and numerically fixed agent primitive that forms the layout on which other agents move in a spatial model. https://helipad.dev/functions/patch/"""
	fixed: bool = True

	#Don't remove from the list; just mark as dead
	def die(self):
		"""Remove the patch from the board, and if `model.spatial()` had been initialized with `offmap=False`, prevent agents from moving to coordinates previously covered by the patch. Patches can be revived with `Patches2D.revive()`. https://helipad.dev/functions/patch/die/"""
		self.dead = True
		if not self.model.patches.offmap:
			for a in self.agentsOn: a.die()
		self.model.doHooks(['baseAgentDie', 'PatchDie'], [self])

	@property
	def neighbors(self) -> list:
		"""A list of adjacent patches. Corner-adjacent patches will be included depending on the value of the `corners` parameter of `model.spatial()`. https://helipad.dev/functions/patch/#neighbors"""
		return [p for p in self.edges.outbound('space', True, obj='agent') if not p.dead]

	def __repr__(self):
		if self.model.patches.geometry == 'geo' and self.name: return f'<Patch {self.name}>'
		return f'<Patch at {self.position[0]},{self.position[1]}>'

#Direction can take an Agent object (corresponding to the endpoint),
#an int (0 for undirected, >0 for agent1 to agent2, and <0 for agent2 to agent1),
#or a boolean (False for undirected, True for agent1 to agent2)
class Edge:
	"""A connection between two agents as part of a network. https://helipad.dev/functions/edge/"""
	def __init__(self, agent1: baseAgent, agent2: baseAgent, kind: str='edge', direction=None, weight=1):
		self.active = True
		self.kind = kind
		self.vertices = (agent1, agent2)
		self.weight = weight
		self.directed = False
		if direction is not None:
			self.directed = True
			if isinstance(direction, int):
				if direction==0: self.directed = False
				elif direction>0: self.startpoint, self.endpoint = (agent1, agent2)
				elif direction<0: self.startpoint, self.endpoint = (agent2, agent1)
			elif isinstance(direction, bool):
				self.directed = direction
				if direction: self.startpoint, self.endpoint = (agent1, agent2)
			elif isinstance(direction, baseAgent):
				if direction not in self.vertices: raise ValueError(ï('Direction must select one of the agents as an endpoint.'))
				self.endpoint = direction
				self.startpoint = agent1 if direction==agent2 else agent2
			else: raise ValueError(ï('Direction must be either int, bool, or agent.'))
		if not self.directed:
			self.endpoint, self.startpoint, self.directed = (None, None, False)

		#Add object to each agent, and to the model
		for agent in self.vertices:
			if kind not in agent.edges: agent.edges[kind] = []
			if self not in agent.edges[kind]: agent.edges[kind].append(self) #Don't add self-links twice

		agent1.model.doHooks('edgeInit', [self, kind, agent1, agent2])

	def __repr__(self):
		pair, arrow = (self.vertices, '—') if not self.directed else ((self.startpoint, self.endpoint), '→')
		return f'<{self.__class__.__name__}: {pair[0].__class__.__name__} {pair[0].id} {arrow} {pair[1].__class__.__name__} {pair[1].id}>'

	def cut(self):
		"""Break the connection between the two agents represented by the edge and remove the `Edge` object from the `edges` property of both agents. https://helipad.dev/functions/edge/cut/"""
		for agent in self.vertices:
			if self in agent.edges[self.kind]: agent.edges[self.kind].remove(self) #Remove from agents
		self.active = False
		self.vertices[0].model.doHooks('edgeCut', [self])

	def partner(self, agent: baseAgent):
		"""When passed an agent, returns the other agent. https://helipad.dev/functions/edge/partner/"""
		if agent==self.vertices[0]: return self.vertices[1]
		elif agent==self.vertices[1]: return self.vertices[0]
		else: raise ValueError(ï('Agent {} is not connected to this edge.').format(agent.id))

	def reassign(self, oldagent: baseAgent, newagent: baseAgent):
		"""Moves a vertex from `oldagent` to `newagent`. https://helipad.dev/functions/edge/reassign/"""
		self.vertices = (self.partner(oldagent), newagent)
		oldagent.edges[self.kind].remove(self)
		newagent.edges[self.kind].append(self)
		newagent.model.doHooks('edgeReassign', [self, oldagent, newagent])

class Stocks:
	"A dict-like interface for agent holdings of registered goods. https://helipad.dev/functions/baseagent/#stocks"
	def __init__(self, breed: str, goodslist: list):
		self.goods = {g:{} for g in goodslist}
		for good, ginfo in goodslist.items():
			for p, fn in ginfo.props.items():
				endow = fn(breed) if callable(fn) else fn
				if endow is None: self.goods[good][p] = 0
				elif isinstance(endow, (tuple, list)): self.goods[good][p] = randint(*endow)
				else: self.goods[good][p] = endow

	def __getitem__(self, key: str):
		if isinstance(key, str): return self.goods[key]['quantity']
		elif isinstance(key, tuple):
			if isinstance(key[1], str): return self.goods[key[0]][key[1]]
			elif key[1] is True: return self.goods[key[0]]
			elif key[1] is False: return self.goods[key]['quantity']
		raise KeyError

	def __setitem__(self, key: str, val):
		if isinstance(key, str): self.goods[key]['quantity'] = val
		elif isinstance(key, tuple) and isinstance(key[1], str): self.goods[key[0]][key[1]] = val
		else: raise KeyError

	def __repr__(self): return {k: g['quantity'] for k,g in self.goods.items()}.__repr__()
	def __iter__(self): return iter({k: g['quantity'] for k,g in self.goods.items()})
	def __next__(self): return next({k: g['quantity'] for k,g in self.goods.items()})
	def __len__(self): return len(self.goods)
	def keys(self): return self.goods.keys()
	def values(self): return [g['quantity'] for g in self.goods.values()]
	def items(self): return [(k, g['quantity']) for k,g in self.goods.items()]

#==================
# CONTAINER CLASSES
#==================

class MultiDict(dict):
	"""A base class for a dict of dicts"""
	def __len__(self):
		return sum(len(a) for a in super().values())

	@property
	def all(self) -> list:
		"""A list of all items contained."""
		agents = []
		for l in self.values(): agents += l
		return agents

class Agents(MultiDict):
	"""Interface for storing and interacting with model agents, organized by primitive. Accessible at `model.agents`. https://helipad.dev/functions/agents/"""
	def __init__(self, model):
		self.model = model
		self.order = 'linear'
		self.edges = ModelEdges(self)
		super().__init__()

	#Allow retrieval by either primitive or agent ID
	def __getitem__(self, val):
		if isinstance(val, int):
			for p in super().values():
				for a in p:
					if a.id==val: return a
		else: return super().__getitem__(val)

	def __contains__(self, val) -> bool:
		if isinstance(val, int): return self[val]
		else: return super().__contains__(val)

	def addPrimitive(self, name: str, class_, plural=None, dflt=50, low=1, high=100, step=1, hidden: bool=False, priority: int=100, order=None):
		"""Register an agent primitive. https://helipad.dev/functions/agents/addprimitive/"""
		if name=='all': raise ValueError(ï('{} is a reserved name. Please choose another.').format(name))
		if not plural: plural = name+'s'
		class_.primitive = name
		self[name] = Primitive(
			class_=class_,
			plural=plural,
			priority=priority,
			order=order,
			breeds=Breeds(self.model, name)
		)
		sort = dict(sorted(self.items(), key=lambda d: d[1].priority))
		self.clear()
		self.update(sort)

		def popget(name, model):
			prim = name.split('_')[1]
			if not model.hasModel: return None
			else: return len(model.agents[prim])
		self.model.params.add('num_'+name, 'Number of '+plural.title(), 'hidden' if hidden else 'slider', dflt=dflt, opts={'low': low, 'high': high, 'step': step} if not hidden else None, setter=self.initialize, getter=popget)

	def removePrimitive(self, name: str):
		"""Removes a previously added primitive. https://helipad.dev/functions/agents/removeprimitive/"""
		del self[name]
		del self.model.params['num_'+name]

	def addBreed(self, name: str, color, prim=None):
		"""Registers an agent breed. https://helipad.dev/functions/agents/addbreed/"""
		if prim is None:
			if len(super().keys()) == 1: prim = next(iter(self.keys()))
			else: raise KeyError(ï('Breed must specify which primitive it belongs to.'))
		return self[prim].breeds.add(name, color)

	#Creates an unweighted and undirected network of a certain density
	def createNetwork(self, density: float, kind: str='edge', prim=None):
		"""Create a random undirected and unweighted network of a certain `density`∈[0,1] among agents of primitive `prim`. https://helipad.dev/functions/agents/createnetwork/"""
		if density < 0 or density > 1: raise ValueError(ï('Network density must take a value between 0 and 1.'))
		from itertools import combinations
		agents = self.all if prim is None else self[prim]
		for c in combinations(agents, 2):
			if np.random.randint(0,100) < density*100:
				c[0].edges.add(c[1], kind)
		return self.network(kind, prim)

	def network(self, kind: str='edge', prim=None, excludePatches: bool=False):
		"""Export the model's graph structure to a `NetworkX` object. Agent breed, primitive, and (if a spatial model) position are stored as metadata. https://helipad.dev/functions/agents/network/"""
		import networkx as nx

		#Have to use DiGraph in order to draw any arrows
		G = nx.DiGraph(name=kind)
		agents = list(self.all) if prim is None else self[prim]
		if excludePatches: agents = [a for a in agents if a.primitive!='patch']
		G.add_nodes_from([(a.id, {'breed': a.breed, 'primitive': a.primitive, 'position': None if a.position is None else list(a.position)}) for a in agents])
		for e in self.edges[kind]:
			if prim is None or (e.vertices[0].primitive==prim and e.vertices[1].primitive==prim): G.add_edge(
				e.startpoint.id if e.directed else e.vertices[0].id,
				e.endpoint.id if e.directed else e.vertices[1].id,
				weight=e.weight, directed=e.directed
			)
		return G

	#Model param redundant, strictly speaking, but it's necessary to make the signature match the setter callback
	def initialize(self, val, prim: str, model=None, force: bool=False):
		"""Create and/or destroy agents to get a population number. This function is used as a setter function for agent population parameters, and also at the beginning of a model to create the initial agent set. https://helipad.dev/functions/agents/initialize/"""
		if not self.model.hasModel and not force: return val

		if 'num_' in prim: prim = prim.split('_')[1] #Because the parameter callback passes num_{prim}
		array = self[prim]
		diff = val - len(array)

		#Add agents
		if diff > 0:
			maxid = max(ids) if (ids := [a.id for a in self.all]) else 0 #Figure out maximum existing ID
			for aId in range(maxid+1, maxid+int(diff)+1):
				breed = self.model.doHooks([prim+'DecideBreed', 'decideBreed'], [aId, self[prim].breeds.keys(), self.model])
				if breed is None: breed = list(self[prim].breeds.keys())[aId%len(self[prim].breeds)]
				if breed not in self[prim].breeds:
					raise ValueError(ï('Breed \'{0}\' is not registered for the \'{1}\' primitive.').format(breed, prim))
				new = self[prim].class_(breed, aId, self.model)
				array.append(new)

		#Remove agents
		elif diff < 0:
			shuffle(array) #Delete agents at random

			#Remove agents, maintaining the proportion between breeds
			n = {x: 0 for x in self[prim].breeds.keys()}
			for a in self[prim]:
				if n[a.breed] < -diff:
					n[a.breed] += 1
					a.die(updateGUI=False)
				else: continue

	#Returns summary statistics on an agent variable at a single point in time
	def summary(self, var: str, prim=None, breed=None, good: bool=False):
		"""Print summary statistics (n, mean, standard deviation, variance, maximum, minimum, and sum) for an agent property. https://helipad.dev/functions/agents/summary/"""
		if prim is None:
			prim = 'agent' if 'agent' in self else next(iter(self))
		agents = self[prim] if breed is None else self[prim][breed]
		if not good: data = pandas.Series([getattr(a, var) for a in agents]) #Pandas gives us nice statistical functions
		else: data = pandas.Series([a.stocks[var] for a in agents])
		stats = {
			'n': data.size,
			'Mean': data.mean(),
			'StDev': data.std(),
			'Variance': data.var(),
			'Maximum': data.max(),
			'Minimum': data.min(),
			'Sum': data.sum()
		}
		for k, v in stats.items():
			print(k+': ', v)

	#Deprecated in Helipad 1.6, remove in Helipad 1.8
	def add(self, *args, **kwargs):
		"""Add a primitive. This method is deprecated; use `model.agents.addPrimitive()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.primitives.add', 'model.agents.addPrimitive'), FutureWarning, 2)
		self.addPrimitive(*args, **kwargs)

	#Deprecated in Helipad 1.6, remove in Helipad 1.8
	def remove(self, name):
		"""Remove a primitive. This method is deprecated; use `model.agents.removePrimitive()` instead."""
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('model.primitives.remove', 'model.agents.removePrimitive'), FutureWarning, 2)
		self.removePrimitive(name)

class Primitive(list):
	"""List-like container for agents of a primitive, plus data defining that primitive. Stored within the `Agents` object. https://helipad.dev/functions/primitive/"""
	def __init__(self, **kwargs):
		for k,v in kwargs.items(): setattr(self, k, v)
		super().__init__()

	def __getitem__(self, val):
		if isinstance(val, str): return [a for a in self if a.breed==val]
		else: return super().__getitem__(val)

class gandb(funcStore):
	"""Base class for breeds and goods containers. Should not be called directly"""
	def __init__(self, model):
		self.model = model

	def add(self, obj: str, name: str, color, **kwargs):
		"""Add an item."""
		if name in self:
			warnings.warn(ï('{0} \'{1}\' already defined. Overriding…').format(obj, name), None, 2)

		cobj = color if isinstance(color, Color) else Color(color)
		cobj2 = cobj.lighten()
		self[name] = Item(color=cobj, color2=cobj2, **kwargs)

		#Make sure the parameter lists keep up with our items
		if obj=='good': paramDict = self.model.params.perGood
		elif obj=='breed': paramDict = {k:v for k,v in self.model.params.perBreed.items() if v.prim==self.primitive}
		for p in paramDict.values(): p.addKey(name)

		return self[name]

	def remove(self, name: str):
		"""Remove an item."""
		if isinstance(name, (list, tuple)): return [self.remove(n) for n in name]
		if name not in self: return False

		#Also delete per-item parameters
		pdict = self.model.params.perBreed if isinstance(self, Breeds) else self.model.params.perGood
		for param in pdict.values():
			del param.value[name]
			if getattr(param, 'elements', False):
				if not isNotebook(): param.elements[name].destroy()
				else: param.elements[name].close()
				del param.elements[name]
		return super().remove(name)

class Breeds(gandb):
	"""Interface for adding and storing data on breeds. Stored within the `Primitive` object. https://helipad.dev/functions/breeds/"""
	def __init__(self, model, primitive: str):
		self.primitive = primitive
		super().__init__(model)

	def add(self, name: str, color):
		"""Register an agent breed. https://helipad.dev/functions/breeds/add/"""
		return super().add('breed', name, color)

class Edges(MultiDict):
	"""Interface for adding and storing connections between agents that define a network. Stored in `Agent.edges`. https://helipad.dev/functions/edges/"""
	def __init__(self, agent: baseAgent):
		self.agent = agent
		super().__init__()

	def add(self, partner: baseAgent, kind: str='edge', direction=None, weight=1):
		"""Create a network connection between the current agent and `partner`. https://helipad.dev/functions/edges/add/"""
		return Edge(self.agent, partner, kind, direction, weight)

	def outbound(self, kind='edge', undirected: bool=False, obj: str='edge'):
		"""Return a list of edges for which the agent is a startpoint. Undirected edges can be excluded or included with `undirected`. https://helipad.dev/functions/edges/outbound/"""
		if obj not in ['agent', 'edge']: raise ValueError(ï('Object must be specified either \'agent\' or \'edge\'.'))
		if kind is None: edges = self.all
		else:
			if kind not in self: return []
			edges = self[kind]
		ob = [edge for edge in edges if edge.startpoint == self.agent or (undirected and not edge.directed)]
		return ob if obj=='edge' else [e.partner(self.agent) for e in ob]

	def inbound(self, kind='edge', undirected: bool=False, obj: str='edge'):
		"""Return a list of edges for which the agent is an endpoint. Undirected edges can be excluded or included with `undirected`. https://helipad.dev/functions/edges/inbound/"""
		if obj not in ['agent', 'edge']: raise ValueError(ï('Object must be specified either \'agent\' or \'edge\'.'))
		if kind is None: edges = self.all
		else:
			if kind not in self: return []
			edges = self[kind]
		ib = [edge for edge in edges if edge.endpoint == self.agent or (undirected and not edge.directed)]
		return ib if obj=='edge' else [e.partner(self.agent) for e in ib]

	def With(self, partner: baseAgent, kind='edge'):
		"""Returns a list of direct connections with `partner`. Does not indicate indirect connections. https://helipad.dev/functions/edges/with/"""
		if kind is not None:
			if kind not in self: return []
			edges = self[kind]
		else: edges = self.all
		return [edge for edge in edges if self.agent in edge.vertices and partner in edge.vertices]

class ModelEdges(MultiDict):
	"""Interface for gathering aggregations of connections between agents. Stored in `model.agent.edges`."""
	def __init__(self, agents: Agents):
		self.agents = agents
		super().__init__()

	def __getitem__(self, val): return [e for e in self.all if e.kind==val]
	def __contains__(self, val) -> bool: return val in list(self.keys())
	def items(self): yield from self._dict.items()
	def values(self): yield from self._dict.values()
	def keys(self):
		ks = []
		for e in self.all:
			if e.kind not in ks: ks.append(e.kind)
		yield from ks
	__iter__ = keys
	def __repr__(self): return self._dict.__repr__()

	@property
	def _dict(self):
		es = {}
		for e in self.all:
			if e.kind not in es: es[e.kind] = []
			if e not in es[e.kind]: es[e.kind].append(e)
		return es

	@property
	def all(self) -> list:
		"""A list of all network edges in the model."""
		es = []
		for a in self.agents.all: es += a.edges.all
		return es



"""
Classes to give agents utility functions, and a base `Utility` class to subclass new utility functions. Built-in utility functions include `CES`, `Leontief`, and `CobbDouglas`.
"""

from abc import ABC, abstractmethod
from helipad.helpers import ï

class Utility(ABC):
	"""A base class defining the methods a utility class must implement. https://helipad.dev/functions/utility/"""

	#Receives an array of goods.
	#Can, but doesn't necessarily have to correspond to the registered goods
	#i.e. you can have an abstract good like real balances
	def __init__(self, goods):
		self.goods = goods
		self.utility = 0

	def consume(self, quantities: dict):
		"""Sets `self.utility` to the value returned from `self.calculate()`. `agent.stocks` can be passed directly to this function. https://helipad.dev/functions/utility/consume/"""
		self.utility = self.calculate(quantities)
		return self.utility

	@abstractmethod
	def calculate(self, quantities: dict):
		"""Calculate the agent's utility on the basis of the array of quantities entered. https://helipad.dev/functions/utility/calculate/"""
		if len(quantities) != len(self.goods): raise KeyError(ï('Quantities argument doesn\'t match initialized list of goods.'))

	@abstractmethod
	def demand(self, budget, prices: dict):
		"""Calculate total demand for the various goods given some income. https://helipad.dev/functions/utility/demand/"""

	@abstractmethod
	def mu(self, quantities: dict):
		"""Calculates the marginal utilities of the goods in question on the basis of the entered quantities. https://helipad.dev/functions/utility/mu/"""
		if len(quantities) != len(self.goods): raise KeyError(ï('Quantities argument doesn\'t match initialized list of goods.'))

	@abstractmethod
	def mrs(self, good1: str, q1, good2: str, q2):
		"""Calculate the marginal rate of substitution between `good1` and `good2` given quantities `q1` and `q2`, respectively. https://helipad.dev/functions/utility/mrs/"""

class CES(Utility):
	"""A constant elasticity of substitution utility function. https://helipad.dev/functions/ces/"""

	#Coefficients should add to 1, but this is not enforced
	#Goods can be a list of goods, or a dict of goods and corresponding coefficients
	def __init__(self, goods, elast):
		if isinstance(goods, dict):
			self.coeffs = goods
			goods = goods.keys()
		else:
			self.coeffs = {g:1 for g in goods}
		super().__init__(goods)
		self.elast = elast

	#Can take agent.goods, if the registered goods correspond
	def calculate(self, quantities):
		super().calculate(quantities)

		#Can't divide by zero in the inner exponent
		#But at σ=0, the utility function becomes a Leontief function in the limit
		#The coefficients don't show up, see https://www.jstor.org/stable/29793581
		if self.elast==0: return min(quantities.values())

		#Can't divide by zero in the outer exponent
		#But at σ=1, the utility function becomes a Cobb-Douglass function
		#See https://yiqianlu.files.wordpress.com/2013/10/ces-functions-and-dixit-stiglitz-formulation.pdf
		elif self.elast==1:
			util = 1
			for g in self.goods:
				util *= quantities[g] ** self.coeffs[g]
			return util

		#Can't raise zero to a negative exponent
		#But as x approaches zero, x^-y approaches infinity
		#So an infinite inner sum means the whole expression becomes zero
		elif (0 in quantities.values()) and self.elast<1: return 0

		#The general utility function
		else:
			util = sum(self.coeffs[g] ** (1/self.elast) * quantities[g] ** ((self.elast-1)/self.elast) for g in self.goods)
			return util ** (self.elast/(self.elast-1))

	def mu(self, quantities):
		super().mu(quantities)

		if self.elast==0:	#Leontief
			mus = {g: 1 if quantities[g] < min(quantities.values()) else 0 for g in self.goods}

		elif self.elast==1:	#Cobb-Douglas
			mus = {}
			for g in self.goods:
				mus[g] = self.coeffs[g] * quantities[g] ** (self.coeffs[g] - 1)
				for g2 in self.goods:
					if g2 != g:
						mus[g] *= quantities[g2] ** self.coeffs[g2]

		else:				#General CES
			coeff = sum(self.coeffs[g] ** (1/self.elast) * quantities[g] ** ((self.elast-1)/self.elast) for g in self.goods)
			coeff = coeff ** (1/(self.elast-1))
			mus = {g: coeff * (self.coeffs[g]/quantities[g]) ** (1/self.elast) for g in self.goods}

		return mus

	#Doesn't depend on any other quantities
	def mrs(self, good1, q1, good2, q2):
		if self.elast==0:
			if q1 < q2: return float('inf')
			elif q1 > q2: return 0
			else: return None #Undefined at the kink in the indifference curve

		#Works for both Cobb-Douglas and the general CES
		else:
			return ((self.coeffs[good1]*q2)/(self.coeffs[good2]*q1)) ** (1/self.elast)

	def demand(self, budget, prices):
		demand = {g:0 for g in self.goods}
		for g in self.goods:
			# Derivation at https://cameronharwick.com/blog/how-to-derive-a-demand-function-from-a-ces-utility-function/
			for h, price in prices.items():
				demand[g] += self.coeffs[h]/self.coeffs[g] * price ** (1-self.elast)
			demand[g] = budget * prices[g] ** (-self.elast) / demand[g]

		return demand

class CobbDouglas(CES):
	"""A Cobb-Douglas utility function. `Goods` can be a list of goods, or a dict of goods and corresponding exponents. https://helipad.dev/functions/cobbdouglas/"""
	def __init__(self, goods):
		if isinstance(goods, list):
			goods = {g:1/len(goods) for g in goods}

		super().__init__(goods, 1)

	@property
	def exponents(self): return self.coeffs

class Leontief(CES):
	"""A Leontief utility function, where utility is equal to the minimum quantity held of any relevant good. https://helipad.dev/functions/leontief/"""
	def __init__(self, goods):
		super().__init__(goods, 0)